#================= MODELING ================================

# shift origin of receiver axis for inversion results
oz_src_land = $(shell echo -0.5*$(dz)*$(src_nz_land) | bc -l)
oz_mute_land = $(shell echo -0.5*$(dz)*$(rec_nz_land) | bc -l)
oz_mute_obn = $(shell echo -0.5*$(dz)*$(rec_nz_obn) | bc -l)
oz_mute_land2 = $(shell echo -0.5*$(dz)*$(rec_nz_land2) | bc -l)

#===========================================================#
#========== modeling from receiver data ====================#
#===========================================================#

#===========================================
#====== Inversion for land data ============
#===========================================

#=== slow velocity
synth_land_inversion_slow: ${B}/PS_sep_inv.x ${D}/lambda2d_const_slow.H ${D}/mu2d_const_slow.H ${D}/rho2d_const_slow.H ${P}/mod2d_land.P ${D}/recs2d_land.H ${D}/srcs2d_ellipse.H ${D}/uxr_land_obs.H ${D}/uzr_land_obs.H ${D}/Pr_land_obs.H
	${B}/PS_sep_inv.x lambda=${D}/lambda2d_const_slow.H mu=${D}/mu2d_const_slow.H rho=${D}/rho2d_const_slow.H par=${P}/mod2d_land.P recs=${D}/recs2d_land.H srcs=${D}/srcs2d_ellipse.H \
	proptype=${RCVR} niter=50 \
	in_ux_data=${D}/uxr_land_obs.H in_uz_data=${D}/uzr_land_obs.H in_P_data=${D}/Pr_land_obs.H \
	rec_uxb=${D}/uxs_land_slow.H rec_uzb=${D}/uzs_land_slow.H

modeling_forward_invresult_land_slow: ${B}/PS_sep_for.x ${D}/lambda2d_const_slow.H ${D}/mu2d_const_slow.H ${D}/rho2d_const_slow.H ${P}/mod2d_land.P ${D}/recs2d_land.H ${D}/srcs2d_ellipse.H ${D}/uxs_land.H ${D}/uzs_land.H
	${B}/PS_sep_for.x lambda=${D}/lambda2d_const_slow.H mu=${D}/mu2d_const_slow.H rho=${D}/rho2d_const_slow.H par=${P}/mod2d_land.P recs=${D}/recs2d_land.H srcs=${D}/srcs2d_ellipse.H \
	proptype=${RCVR} adj=0 splitdisp=0 \
	in_ux_src=${D}/uxs_land_slow.H in_uz_src=${D}/uzs_land_slow.H \
	rec_ux=${D}/uxr_land_inv_slow.H rec_uz=${D}/uzr_land_inv_slow.H \
	rec_P=${D}/Pr_land_inv_slow.H rec_S=${D}/Sr_land_inv_slow.H
	Reverse < ${D}/Sr_land_inv_slow.H > ${D}/stmp.H
	Mv ${D}/stmp.H ${D}/Sr_land_inv_slow.H
	echo "d1=${dz}" >> ${D}/Pr_land_inv_slow.H
	echo "d1=${dz}" >> ${D}/Sr_land_inv_slow.H
	echo "d1=${dz}" >> ${D}/uxr_land_inv_slow.H
	echo "d1=${dz}" >> ${D}/uzr_land_inv_slow.H
	echo "o1=${oz_mute_land}" >> ${D}/Pr_land_inv_slow.H
	echo "o1=${oz_mute_land}" >> ${D}/Sr_land_inv_slow.H
	echo "o1=${oz_mute_land}" >> ${D}/uxr_land_inv_slow.H
	echo "o1=${oz_mute_land}" >> ${D}/uzr_land_inv_slow.H

#=== correct velocity
synth_land_inversion: ${B}/PS_sep_inv.x ${D}/lambda2d_const.H ${D}/mu2d_const.H ${D}/rho2d_const.H ${P}/mod2d_land.P ${D}/recs2d_land.H ${D}/srcs2d_ellipse.H ${D}/uxr_land_obs.H ${D}/uzr_land_obs.H ${D}/Pr_land_obs.H
	${B}/PS_sep_inv.x lambda=${D}/lambda2d_const.H mu=${D}/mu2d_const.H rho=${D}/rho2d_const.H par=${P}/mod2d_land.P recs=${D}/recs2d_land.H srcs=${D}/srcs2d_ellipse.H \
	proptype=${RCVR} niter=50 \
	in_ux_data=${D}/uxr_land_obs.H in_uz_data=${D}/uzr_land_obs.H in_P_data=${D}/Pr_land_obs.H \
	rec_uxb=${D}/uxs_land.H rec_uzb=${D}/uzs_land.H

modeling_forward_invresult_land: ${B}/PS_sep_for.x ${D}/lambda2d_const.H ${D}/mu2d_const.H ${D}/rho2d_const.H ${P}/mod2d_land.P ${D}/recs2d_land.H ${D}/srcs2d_ellipse.H ${D}/uxs_land.H ${D}/uzs_land.H
	${B}/PS_sep_for.x lambda=${D}/lambda2d_const.H mu=${D}/mu2d_const.H rho=${D}/rho2d_const.H par=${P}/mod2d_land.P recs=${D}/recs2d_land.H srcs=${D}/srcs2d_ellipse.H \
	proptype=${RCVR} adj=0 splitdisp=0 \
	in_ux_src=${D}/uxs_land.H in_uz_src=${D}/uzs_land.H \
	rec_ux=${D}/uxr_land_inv.H rec_uz=${D}/uzr_land_inv.H \
	rec_P=${D}/Pr_land_inv.H rec_S=${D}/Sr_land_inv.H
	Reverse < ${D}/Sr_land_inv.H > ${D}/stmp.H
	Mv ${D}/stmp.H ${D}/Sr_land_inv.H
	echo "d1=${dz}" >> ${D}/Pr_land_inv.H
	echo "d1=${dz}" >> ${D}/Sr_land_inv.H
	echo "d1=${dz}" >> ${D}/uxr_land_inv.H
	echo "d1=${dz}" >> ${D}/uzr_land_inv.H
	echo "d1=${dz}" >> ${D}/uxs_land.H
	echo "d1=${dz}" >> ${D}/uzs_land.H
	echo "o1=${oz_mute_land}" >> ${D}/Pr_land_inv.H
	echo "o1=${oz_mute_land}" >> ${D}/Sr_land_inv.H
	echo "o1=${oz_mute_land}" >> ${D}/uxr_land_inv.H
	echo "o1=${oz_mute_land}" >> ${D}/uzr_land_inv.H
	echo "o1=${oz_src_land}" >> ${D}/uxs_land.H
	echo "o1=${oz_src_land}" >> ${D}/uzs_land.H

#=== fast velocity
synth_land_inversion_fast: ${B}/PS_sep_inv.x ${D}/lambda2d_const_fast.H ${D}/mu2d_const_fast.H ${D}/rho2d_const_fast.H ${P}/mod2d_land.P ${D}/recs2d_land.H ${D}/srcs2d_ellipse.H ${D}/uxr_land_obs.H ${D}/uzr_land_obs.H ${D}/Pr_land_obs.H
	${B}/PS_sep_inv.x lambda=${D}/lambda2d_const_fast.H mu=${D}/mu2d_const_fast.H rho=${D}/rho2d_const_fast.H par=${P}/mod2d_land.P recs=${D}/recs2d_land.H srcs=${D}/srcs2d_ellipse.H \
	proptype=${RCVR} niter=50 \
	in_ux_data=${D}/uxr_land_obs.H in_uz_data=${D}/uzr_land_obs.H in_P_data=${D}/Pr_land_obs.H \
	rec_uxb=${D}/uxs_land_fast.H rec_uzb=${D}/uzs_land_fast.H

modeling_forward_invresult_land_fast: ${B}/PS_sep_for.x ${D}/lambda2d_const_fast.H ${D}/mu2d_const_fast.H ${D}/rho2d_const_fast.H ${P}/mod2d_land.P ${D}/recs2d_land.H ${D}/srcs2d_ellipse.H ${D}/uxs_land.H ${D}/uzs_land.H
	${B}/PS_sep_for.x lambda=${D}/lambda2d_const_fast.H mu=${D}/mu2d_const_fast.H rho=${D}/rho2d_const_fast.H par=${P}/mod2d_land.P recs=${D}/recs2d_land.H srcs=${D}/srcs2d_ellipse.H \
	proptype=${RCVR} adj=0 splitdisp=0 \
	in_ux_src=${D}/uxs_land_fast.H in_uz_src=${D}/uzs_land_fast.H \
	rec_ux=${D}/uxr_land_inv_fast.H rec_uz=${D}/uzr_land_inv_fast.H \
	rec_P=${D}/Pr_land_inv_fast.H rec_S=${D}/Sr_land_inv_fast.H
	Reverse < ${D}/Sr_land_inv_fast.H > ${D}/stmp.H
	Mv ${D}/stmp.H ${D}/Sr_land_inv_fast.H
	echo "d1=${dz}" >> ${D}/Pr_land_inv_fast.H
	echo "d1=${dz}" >> ${D}/Sr_land_inv_fast.H
	echo "d1=${dz}" >> ${D}/uxr_land_inv_fast.H
	echo "d1=${dz}" >> ${D}/uzr_land_inv_fast.H
	echo "o1=${oz_mute_land}" >> ${D}/Pr_land_inv_fast.H
	echo "o1=${oz_mute_land}" >> ${D}/Sr_land_inv_fast.H
	echo "o1=${oz_mute_land}" >> ${D}/uxr_land_inv_fast.H
	echo "o1=${oz_mute_land}" >> ${D}/uzr_land_inv_fast.H

#===========================================
#====== Inversion for OBN data ============
#===========================================

#=== correct velocity
synth_obn_inversion: ${B}/PS_sep_inv.x ${D}/lambda2d_const.H ${D}/mu2d_const.H ${D}/rho2d_const.H ${P}/mod2d_obn.P ${D}/recs2d_obn.H ${D}/srcs2d_obn.H ${D}/uxr_obn_obs.H ${D}/uzr_obn_obs.H ${D}/Pr_obn_obs.H
	${B}/PS_sep_inv.x lambda=${D}/lambda2d_const.H mu=${D}/mu2d_const.H rho=${D}/rho2d_const.H par=${P}/mod2d_obn.P recs=${D}/recs2d_obn.H srcs=${D}/srcs2d_obn.H \
	proptype=${RCVR} niter=20 \
	in_ux_data=${D}/uxr_obn_obs.H in_uz_data=${D}/uzr_obn_obs.H in_P_data=${D}/Pr_obn_obs.H \
	rec_uxb=${D}/uxs_obn.H rec_uzb=${D}/uzs_obn.H

modeling_forward_invresult_obn: ${B}/PS_sep_for.x ${D}/lambda2d_const.H ${D}/mu2d_const.H ${D}/rho2d_const.H ${P}/mod2d_obn.P ${D}/recs2d_obn.H ${D}/srcs2d_obn.H ${D}/uxs_obn.H ${D}/uzs_obn.H
	${B}/PS_sep_for.x lambda=${D}/lambda2d_const.H mu=${D}/mu2d_const.H rho=${D}/rho2d_const.H par=${P}/mod2d_obn.P recs=${D}/recs2d_obn.H srcs=${D}/srcs2d_obn.H \
	proptype=${RCVR} adj=0 splitdisp=0 \
	in_ux_src=${D}/uxs_obn.H in_uz_src=${D}/uzs_obn.H \
	rec_ux=${D}/uxr_obn_inv.H rec_uz=${D}/uzr_obn_inv.H \
	rec_P=${D}/Pr_obn_inv.H rec_S=${D}/Sr_obn_inv.H
	Reverse < ${D}/Sr_obn_inv.H > ${D}/stmp.H
	Mv ${D}/stmp.H ${D}/Sr_obn_inv.H
	echo "d1=${dz}" >> ${D}/Pr_obn_inv.H
	echo "d1=${dz}" >> ${D}/Sr_obn_inv.H
	echo "d1=${dz}" >> ${D}/uxr_obn_inv.H
	echo "d1=${dz}" >> ${D}/uzr_obn_inv.H
	echo "o1=${oz_mute_obn}" >> ${D}/Pr_obn_inv.H
	echo "o1=${oz_mute_obn}" >> ${D}/Sr_obn_inv.H
	echo "o1=${oz_mute_obn}" >> ${D}/uxr_obn_inv.H
	echo "o1=${oz_mute_obn}" >> ${D}/uzr_obn_inv.H

#===========================================
#====== Inversion for silly land data ============
#===========================================

#=== slow velocity
synth_land2_inversion_slow: ${B}/PS_sep_inv.x ${D}/lambda2d_const_slow.H ${D}/mu2d_const_slow.H ${D}/rho2d_const_slow.H ${P}/mod2d_land2.P ${D}/recs2d_land2.H ${D}/srcs2d_land2.H ${D}/uxr_land2_obs.H ${D}/uzr_land2_obs.H ${D}/Pr_land2_obs.H
	${B}/PS_sep_inv.x lambda=${D}/lambda2d_const_slow.H mu=${D}/mu2d_const_slow.H rho=${D}/rho2d_const_slow.H par=${P}/mod2d_land2.P recs=${D}/recs2d_land2.H srcs=${D}/srcs2d_land2.H \
	proptype=${RCVR} niter=30 \
	in_ux_data=${D}/uxr_land2_obs.H in_uz_data=${D}/uzr_land2_obs.H in_P_data=${D}/Pr_land2_obs.H \
	rec_uxb=${D}/uxs_land2_slow.H rec_uzb=${D}/uzs_land2_slow.H

modeling_forward_invresult_land2_slow: ${B}/PS_sep_for.x ${D}/lambda2d_const_slow.H ${D}/mu2d_const_slow.H ${D}/rho2d_const_slow.H ${P}/mod2d_land2.P ${D}/recs2d_land2.H ${D}/srcs2d_land2.H ${D}/uxs_land2.H ${D}/uzs_land2.H
	${B}/PS_sep_for.x lambda=${D}/lambda2d_const_slow.H mu=${D}/mu2d_const_slow.H rho=${D}/rho2d_const_slow.H par=${P}/mod2d_land2.P recs=${D}/recs2d_land2.H srcs=${D}/srcs2d_land2.H \
	proptype=${RCVR} adj=0 splitdisp=0 \
	in_ux_src=${D}/uxs_land2_slow.H in_uz_src=${D}/uzs_land2_slow.H \
	rec_ux=${D}/uxr_land2_inv_slow.H rec_uz=${D}/uzr_land2_inv_slow.H \
	rec_P=${D}/Pr_land2_inv_slow.H rec_S=${D}/Sr_land2_inv_slow.H
	Reverse < ${D}/Sr_land2_inv_slow.H > ${D}/stmp.H
	Mv ${D}/stmp.H ${D}/Sr_land2_inv_slow.H
	echo "d1=${dz}" >> ${D}/Pr_land2_inv_slow.H
	echo "d1=${dz}" >> ${D}/Sr_land2_inv_slow.H
	echo "d1=${dz}" >> ${D}/uxr_land2_inv_slow.H
	echo "d1=${dz}" >> ${D}/uzr_land2_inv_slow.H
	echo "o1=${oz_mute_land2}" >> ${D}/Pr_land2_inv_slow.H
	echo "o1=${oz_mute_land2}" >> ${D}/Sr_land2_inv_slow.H
	echo "o1=${oz_mute_land2}" >> ${D}/uxr_land2_inv_slow.H
	echo "o1=${oz_mute_land2}" >> ${D}/uzr_land2_inv_slow.H

#=== correct velocity
synth_land2_inversion: ${B}/PS_sep_inv.x ${D}/lambda2d_const.H ${D}/mu2d_const.H ${D}/rho2d_const.H ${P}/mod2d_land2.P ${D}/recs2d_land2.H ${D}/srcs2d_land2.H ${D}/uxr_land2_obs.H ${D}/uzr_land2_obs.H ${D}/Pr_land2_obs.H
	${B}/PS_sep_inv.x lambda=${D}/lambda2d_const.H mu=${D}/mu2d_const.H rho=${D}/rho2d_const.H par=${P}/mod2d_land2.P recs=${D}/recs2d_land2.H srcs=${D}/srcs2d_land2.H \
	proptype=${RCVR} niter=30 \
	in_ux_data=${D}/uxr_land2_obs.H in_uz_data=${D}/uzr_land2_obs.H in_P_data=${D}/Pr_land2_obs.H \
	rec_uxb=${D}/uxs_land2.H rec_uzb=${D}/uzs_land2.H

modeling_forward_invresult_land2: ${B}/PS_sep_for.x ${D}/lambda2d_const.H ${D}/mu2d_const.H ${D}/rho2d_const.H ${P}/mod2d_land2.P ${D}/recs2d_land2.H ${D}/srcs2d_land2.H ${D}/uxs_land2.H ${D}/uzs_land2.H
	${B}/PS_sep_for.x lambda=${D}/lambda2d_const.H mu=${D}/mu2d_const.H rho=${D}/rho2d_const.H par=${P}/mod2d_land2.P recs=${D}/recs2d_land2.H srcs=${D}/srcs2d_land2.H \
	proptype=${RCVR} adj=0 splitdisp=0 \
	in_ux_src=${D}/uxs_land2.H in_uz_src=${D}/uzs_land2.H \
	rec_ux=${D}/uxr_land2_inv.H rec_uz=${D}/uzr_land2_inv.H \
	rec_P=${D}/Pr_land2_inv.H rec_S=${D}/Sr_land2_inv.H
	Reverse < ${D}/Sr_land2_inv.H > ${D}/stmp.H
	Mv ${D}/stmp.H ${D}/Sr_land2_inv.H
	echo "d1=${dz}" >> ${D}/Pr_land2_inv.H
	echo "d1=${dz}" >> ${D}/Sr_land2_inv.H
	echo "d1=${dz}" >> ${D}/uxr_land2_inv.H
	echo "d1=${dz}" >> ${D}/uzr_land2_inv.H
	echo "o1=${oz_mute_land2}" >> ${D}/Pr_land2_inv.H
	echo "o1=${oz_mute_land2}" >> ${D}/Sr_land2_inv.H
	echo "o1=${oz_mute_land2}" >> ${D}/uxr_land2_inv.H
	echo "o1=${oz_mute_land2}" >> ${D}/uzr_land2_inv.H

#=== fast velocity
synth_land2_inversion_fast: ${B}/PS_sep_inv.x ${D}/lambda2d_const_fast.H ${D}/mu2d_const_fast.H ${D}/rho2d_const_fast.H ${P}/mod2d_land2.P ${D}/recs2d_land2.H ${D}/srcs2d_land2.H ${D}/uxr_land2_obs.H ${D}/uzr_land2_obs.H ${D}/Pr_land2_obs.H
	${B}/PS_sep_inv.x lambda=${D}/lambda2d_const_fast.H mu=${D}/mu2d_const_fast.H rho=${D}/rho2d_const_fast.H par=${P}/mod2d_land2.P recs=${D}/recs2d_land2.H srcs=${D}/srcs2d_land2.H \
	proptype=${RCVR} niter=30 \
	in_ux_data=${D}/uxr_land2_obs.H in_uz_data=${D}/uzr_land2_obs.H in_P_data=${D}/Pr_land2_obs.H \
	rec_uxb=${D}/uxs_land2_fast.H rec_uzb=${D}/uzs_land2_fast.H

modeling_forward_invresult_land2_fast: ${B}/PS_sep_for.x ${D}/lambda2d_const_fast.H ${D}/mu2d_const_fast.H ${D}/rho2d_const_fast.H ${P}/mod2d_land2.P ${D}/recs2d_land2.H ${D}/srcs2d_land2.H ${D}/uxs_land2.H ${D}/uzs_land2.H
	${B}/PS_sep_for.x lambda=${D}/lambda2d_const_fast.H mu=${D}/mu2d_const_fast.H rho=${D}/rho2d_const_fast.H par=${P}/mod2d_land2.P recs=${D}/recs2d_land2.H srcs=${D}/srcs2d_land2.H \
	proptype=${RCVR} adj=0 splitdisp=0 \
	in_ux_src=${D}/uxs_land2_fast.H in_uz_src=${D}/uzs_land2_fast.H \
	rec_ux=${D}/uxr_land2_inv_fast.H rec_uz=${D}/uzr_land2_inv_fast.H \
	rec_P=${D}/Pr_land2_inv_fast.H rec_S=${D}/Sr_land2_inv_fast.H
	Reverse < ${D}/Sr_land2_inv_fast.H > ${D}/stmp.H
	Mv ${D}/stmp.H ${D}/Sr_land2_inv_fast.H
	echo "d1=${dz}" >> ${D}/Pr_land2_inv_fast.H
	echo "d1=${dz}" >> ${D}/Sr_land2_inv_fast.H
	echo "d1=${dz}" >> ${D}/uxr_land2_inv_fast.H
	echo "d1=${dz}" >> ${D}/uzr_land2_inv_fast.H
	echo "o1=${oz_mute_land2}" >> ${D}/Pr_land2_inv_fast.H
	echo "o1=${oz_mute_land2}" >> ${D}/Sr_land2_inv_fast.H
	echo "o1=${oz_mute_land2}" >> ${D}/uxr_land2_inv_fast.H
	echo "o1=${oz_mute_land2}" >> ${D}/uzr_land2_inv_fast.H
