program PS_sep_inv
	use sep
	use distrib_type_mod
	use model_mod
	use io_mod
	use solver_smp_mod
	use cgstep_mod

	implicit none

	type (model_type)			:: model_s
	type (distrib_type)			:: distrib_s
	type (io_type)				:: io_s_main
	integer, external			:: omp_get_thread_num, omp_get_num_threads
	logical						:: logic
	integer						:: niter
	


	call sep_init (SOURCE)


	!read model parameters
	write (0,*) "MODEL INIT"
	call model_init (model_s, "lambda")

	!output init
	write (0,*) "IO INIT"
	call io_init (model_s, io_s_main)

	!allocate,read, and distribute
	write (0,*) "CREATE ARRAYS"
	call create_arrays (distrib_s, model_s)
	

	!stupid global variable init
	write (0,*) "STUPID GLOBAL VARIABLE INIT"
	call global_init (model_s, distrib_s, io_s_main)

	! read receiver data
	call load_recs ("in_ux_data", io_s_main%in_ux_data, model_s, distrib_s, uxr_i)
	call load_recs ("in_uz_data", io_s_main%in_uz_data, model_s, distrib_s, uzr_i)
	call load_recs ("in_P_data",  io_s_main%in_P_data,  model_s, distrib_s, Pr_i)

	write (0,*) "SET TIMERS"
	logic = init_sep_timers  ()
	logic = setup_next_timer ("V4",tm_v4)
	logic = setup_next_timer ("U4",tm_u4)
	logic = setup_next_timer ("Disp",tm_se)
	logic = setup_next_timer ("DispSpace",tm_space)
	logic = setup_next_timer ("Sponge",tm_sp)
	logic = setup_next_timer ("Write",tm_write)
	logic = setup_next_timer ("Read",tm_read)

	call from_param ("niter", niter, 10)

	!call solver_smp(m=distrib_s%srcmodel, d=distrib_s%recdata, Fop=L_oper, stepper=cgstep, niter=niter, verb=.true.)
	call solver_smp(m=distrib_s%srcmodel, d=distrib_s%recdata, Fop=L_oper, stepper=cgstep, niter=niter, Wop=W_oper, verb=.true.) ! weight the data space

	! write source model
	call save_srcs ("rec_uxb", io_s_main%rec_uxb, model_s, distrib_s, uxs_i)
	call save_srcs ("rec_uzb", io_s_main%rec_uzb, model_s, distrib_s, uzs_i)

	call print_timers()

end program
