module elas2d_io_type_mod


  type io_type
    logical :: snap_P, snap_S
    logical :: snap_ux, snap_uz
    logical :: snap_vx, snap_vz
	logical :: snap_txx, snap_txz, snap_tzz
    logical :: rec_P, rec_S, rec_ux, rec_uz, rec_vx, rec_vz, rec_txx, rec_txz, rec_tzz
	logical :: rec_Upx, rec_Upz, rec_Usx, rec_Usz, rec_uzdz
	logical	:: in_P, in_vx, in_vz
    integer    :: jsnap,jrec
  end type


end module

module elas2d_io_mod
  use sep
  use elas2d_distrib_type_mod
  use elas2d_io_type_mod
  use elas2d_model_type_mod
  use elas2d_thread_type_mod
  implicit none
  real, pointer,private :: slice2d(:,:),slice2d_s(:,:)
  real, pointer,private :: recs(:) ! temporary contigous array for receiver data before writing it out to file.

  integer,private,save	:: surf_type
  contains
  
  subroutine create_snap_file(model_s,io_s,file,snap_do)
    type( model_type)  :: model_s 
    type(io_type)      :: io_s
    logical            :: snap_do
    character(len=*)   :: file
    character(len=256) :: temp
    integer ::n_t

    call from_param(file,temp,"none")
    snap_do=.false.
    n_t=ceiling(real(model_s%nt)/real(io_s%jsnap))
    if(temp(1:4)/="none") then
      snap_do=.true.
      call sep_put_data_axis_par (file, 1, model_s%nx-model_s%nsponge_x*2, 0., model_s%dx (1), "x")
      call sep_put_data_axis_par (file, 2, model_s%nz-model_s%nsponge_z*2, 0., model_s%dz (1), "z")
      call sep_put_data_axis_par (file, 3, n_t,							   0., model_s%dt*io_s%jsnap, "time")
    end if
  end subroutine

  subroutine create_rec_file(model_s,io_s,file,rec_do)
    type( model_type)  :: model_s 
    type(io_type)      :: io_s
    logical            :: rec_do
    character(len=*)   :: file
    character(len=256) :: temp
    integer ::n_t

    call from_param (file,temp,"none")
    rec_do = .false.
    n_t = ceiling (real (model_s%nt) / real (io_s%jrec))
    if (temp (1:4) /= "none") then
      rec_do = .true.
      call sep_put_data_axis_par (file, 1, size (model_s%recs,2), 0., 1., "receiver")
      call sep_put_data_axis_par (file, 2, n_t, 0., model_s%dt*io_s%jrec, "time")
    end if
  end subroutine

  subroutine create_input_file (file,in_do)
    logical            :: in_do
    character(len=*)   :: file
    character(len=256) :: temp

    call from_param(file,temp,"none")
    in_do=.false.
    if(temp(1:4)/="none") then
      in_do=.true.
    end if
  end subroutine

  subroutine io_init(model_s,io_s)
    type(io_type)		:: io_s
    type(model_type)	:: model_s

    call from_param("surf_type",surf_type,1)
    call from_param("jrec",io_s%jrec,1)
    call from_param("jsnap",io_s%jsnap,10)
    call create_snap_file(model_s,io_s,"snap_P",io_s%snap_P)
    call create_snap_file(model_s,io_s,"snap_S",io_s%snap_S)
    call create_snap_file(model_s,io_s,"snap_ux",io_s%snap_ux)
    call create_snap_file(model_s,io_s,"snap_uz",io_s%snap_uz)
    call create_snap_file(model_s,io_s,"snap_vx",io_s%snap_vx)
    call create_snap_file(model_s,io_s,"snap_vz",io_s%snap_vz)
    call create_snap_file(model_s,io_s,"snap_txx",io_s%snap_txx)
    call create_snap_file(model_s,io_s,"snap_tzz",io_s%snap_tzz)
    call create_snap_file(model_s,io_s,"snap_txz",io_s%snap_txz)

    call create_rec_file(model_s,io_s,"rec_P",io_s%rec_P)
    call create_rec_file(model_s,io_s,"rec_S",io_s%rec_S)
    call create_rec_file(model_s,io_s,"rec_ux",io_s%rec_ux)
    call create_rec_file(model_s,io_s,"rec_uz",io_s%rec_uz)
    call create_rec_file(model_s,io_s,"rec_vx",io_s%rec_vx)
    call create_rec_file(model_s,io_s,"rec_vz",io_s%rec_vz)
    call create_rec_file(model_s,io_s,"rec_txx",io_s%rec_txx)
    call create_rec_file(model_s,io_s,"rec_tzz",io_s%rec_tzz)
    call create_rec_file(model_s,io_s,"rec_txz",io_s%rec_txz)
    call create_rec_file(model_s,io_s,"rec_Upx",io_s%rec_Upx)
    call create_rec_file(model_s,io_s,"rec_Upz",io_s%rec_Upz)
    call create_rec_file(model_s,io_s,"rec_Usx",io_s%rec_Usx)
    call create_rec_file(model_s,io_s,"rec_Usz",io_s%rec_Usz)
    call create_rec_file(model_s,io_s,"rec_uzdz",io_s%rec_uzdz)

	call create_input_file ("in_P", io_s%in_P)
	call create_input_file ("in_vx", io_s%in_vx)
	call create_input_file ("in_vz", io_s%in_vz)

    allocate (slice2d_s (model_s%nx, model_s%nz))
    allocate (slice2d   (model_s%nx - model_s%nsponge_x * 2, model_s%nz - model_s%nsponge_z * 2))
    allocate (recs (size (model_s%recs, 2)))
  end subroutine


  subroutine write_out_files(it,th,m_s,d_s,io_s)
    type(thread_type) :: th
    type(model_type)   :: m_s
    type(distrib_type) :: d_s(:)
    type(io_type)      :: io_s
    integer            :: it


    if(io_s%jsnap ==1 .or. mod(it,io_s%jsnap)==1) then

      if(io_s%snap_P) call create_snap("snap_P",th,m_s,d_s,P_i)
      if(io_s%snap_S) call create_snap("snap_S",th,m_s,d_s,S_i)
      if(io_s%snap_ux) call create_snap("snap_ux",th,m_s,d_s,ux_i)
      if(io_s%snap_uz) call create_snap("snap_uz",th,m_s,d_s,uz_i)
      if(io_s%snap_vx) call create_snap("snap_vx",th,m_s,d_s,vx_i)
      if(io_s%snap_vz) call create_snap("snap_vz",th,m_s,d_s,vz_i)
      if(io_s%snap_txx) call create_snap("snap_txx",th,m_s,d_s,txx_i)
      if(io_s%snap_tzz) call create_snap("snap_tzz",th,m_s,d_s,tzz_i)
      if(io_s%snap_txz) call create_snap("snap_txz",th,m_s,d_s,txz_i)
    end if
    

    if(io_s%jrec ==1 .or. mod(it,io_s%jrec)==1) then
      if(io_s%rec_P) call create_recs("rec_P",th,m_s,d_s,P_i)
      if(io_s%rec_S) call create_recs("rec_S",th,m_s,d_s,S_i)
      if(io_s%rec_ux) call create_recs("rec_ux",th,m_s,d_s,ux_i)
      if(io_s%rec_uz) call create_recs("rec_uz",th,m_s,d_s,uz_i)
      if(io_s%rec_vx) call create_recs("rec_vx",th,m_s,d_s,vx_i)
      if(io_s%rec_vz) call create_recs("rec_vz",th,m_s,d_s,vz_i)
      if(io_s%rec_txx) call create_recs("rec_txx",th,m_s,d_s,txx_i)
      if(io_s%rec_tzz) call create_recs("rec_tzz",th,m_s,d_s,tzz_i)
      if(io_s%rec_txz) call create_recs("rec_txz",th,m_s,d_s,txz_i)
      if(io_s%rec_Upx) call create_recs("rec_Upx",th,m_s,d_s,Upx_i)
      if(io_s%rec_Upz) call create_recs("rec_Upz",th,m_s,d_s,Upz_i)
      if(io_s%rec_Usx) call create_recs("rec_Usx",th,m_s,d_s,Usx_i)
      if(io_s%rec_Usz) call create_recs("rec_Usz",th,m_s,d_s,Usz_i)
      if(io_s%rec_uzdz) call create_recs("rec_uzdz",th,m_s,d_s,uzdz_i)
    end if
  end subroutine

subroutine create_recs(tag,th,m_s,d_s,iarray)
	character(len=*)	:: tag
	integer				:: ith
	type(model_type)	:: m_s
	type(distrib_type)	:: d_s (:)
	type(thread_type)	:: th
	integer				:: i
	integer,external	:: srite
	integer				:: iarray, iloc (2), irec

	ith = th%ith

	if (ith == 1) then

		do i = 1, size (recs)
			ith	 = m_s%rec_ith (i)
			irec = m_s%rec_i (i)
			iloc = d_s(ith)%recs (:, irec)
			recs(i) = d_s(ith)%all (iloc (1), iloc (2), iarray)
		end do
		if (4 * size (recs) /= srite (tag, recs, size(recs) * 4)) then
			write (0, *) trim (tag)
			call seperr ("trouble writing out recs")
		end if

	end if
end subroutine

subroutine read_recs (tag, in_do, ith, m_s, d_s, iarray1, iarray2)
	character(len=*)	:: tag
	logical				:: in_do
	integer				:: ith
	type(model_type)	:: m_s
	type(distrib_type)	:: d_s (:)
	integer				:: i
	integer,external	:: sreed
	integer				:: iarray1, iarray2, iloc (2), irec
    character(len=256) :: temp

	if (ith == 1 .and. in_do) then

		if (4 * size (recs) /= sreed (tag, recs, size(recs) * 4)) then
			write (0, *) trim (tag)
			call seperr ("trouble reading in recs")
		end if

		do i = 1, size (recs)
			ith	 = m_s%rec_ith (i)
			irec = m_s%rec_i (i)
			iloc = d_s(ith)%recs (:, irec)

			!inject velocity from geophones - only if the geophone location has shear strength (not in the water layer)
			if (iarray1 == vx_i .or. iarray1 == vz_i) then !inject velocity from geophones
				if (d_s(ith)%mu (iloc (1), iloc (2)) .ne. 0.) then
					d_s(ith)%all (iloc (1), iloc (2), iarray1) = d_s(ith)%all (iloc (1), iloc (2), iarray1) + recs (i)
																	   ! / (d_s(ith)%dx (iloc(1)) * d_s(ith)%dz (iloc(2)))
				endif
			else !inject pressure into stress fields - only if the hydrophone location has NO shear strength (in the water layer)
				if ( abs(d_s(ith)%mu (iloc (1), iloc (2))) .lt. 1e-4) then
					d_s(ith)%all (iloc (1), iloc (2), iarray1) = d_s(ith)%all (iloc (1), iloc (2), iarray1) + 0.5 * recs (i)
																		! * m_s%dt / (d_s(ith)%dx (iloc(1)) * d_s(ith)%dz (iloc(2)))

					d_s(ith)%all (iloc (1), iloc (2), iarray2) = d_s(ith)%all (iloc (1), iloc (2), iarray2) + 0.5 * recs (i)
																		! * m_s%dt / (d_s(ith)%dx (iloc(1)) * d_s(ith)%dz (iloc(2)))
				endif
			endif
		end do

	endif

end subroutine

subroutine create_snap (tag,th,m_s,d_s,iarray)
	character (len=*)   :: tag
	type (model_type)   :: m_s
	type (distrib_type) :: d_s(:)
	type (thread_type)  :: th
	integer             :: ix, iz, ierr
	integer,external    :: srite
	integer             :: iarray, ibz, iez, ibx, iex, ith
	
	ith=th%ith
	
	slice2d = 0.
	slice2d_s = 0.
	
	if (ith == 1) then
		do ith = 1, th%nth
			ibz = d_s (ith)%fz + 1
			iez = d_s (ith)%fz + d_s (ith)%nz
			ibx = d_s (ith)%fx + 1
			iex = d_s (ith)%fx + d_s (ith)%nx
			
			slice2d_s (ibx:iex, ibz:iez) = d_s (ith)%all (:, :, iarray)
		end do

		do iz = 1, size (slice2d, 2)
			do ix = 1, size (slice2d, 1)
				slice2d (ix, iz) = slice2d_s (ix + m_s%nsponge_x, iz + m_s%nsponge_z)
			end do 
		end do 

		if (4 * size (slice2d) /= srite (tag, slice2d, size (slice2d) * 4)) then
			write (0,*), "=slice tag=", trim (tag)
			call seperr ("trouble writing out slice")
		end if
	end if
end subroutine


subroutine create_arrays (d_s, m_s)
	type(distrib_type)	:: d_s(:)
	type(model_type)	:: m_s
	real				:: dmin, dmax, wavel, vmin, vmax, min_vs, vp, vs
	integer				:: iz, ix, nmax, i
	integer				:: nwavel_samples

	do i = 1, size (d_s)
		allocate (d_s (i)%rho       (d_s (i)%nx, d_s (i)%nz))
		allocate (d_s (i)%lambda    (d_s (i)%nx, d_s (i)%nz))
		allocate (d_s (i)%mu        (d_s (i)%nx, d_s (i)%nz))
	end do

	call read_distribute_rho_omp ("rho", m_s, d_s)
	call read_distribute_mu_omp  ("mu",  m_s, d_s)
	call read_distribute_lam_omp ("lambda",  m_s, d_s)
	
    call from_param("min_vs", min_vs, 500.)

	do i = 1, size (d_s)
		vp = sqrt ((d_s (i)%lambda (d_s (i)%nx / 2, d_s (i)%nz / 2) + 2 * d_s (i)%mu (d_s (i)%nx / 2, d_s (i)%nz / 2)) * d_s (i)%rho (d_s (i)%nx / 2, d_s (i)%nz / 2))
		vmin = vp
		vmax = vp
		do iz = m_s%nsponge_z+1, d_s(i)%nz - m_s%nsponge_z
			do ix = m_s%nsponge_x+1, d_s(i)%nx - m_s%nsponge_x
				vp = sqrt ((d_s (i)%lambda (ix,iz) + 2 * d_s (i)%mu (ix,iz)) * d_s (i)%rho (ix,iz))
				vs = sqrt (d_s (i)%mu (ix,iz) * d_s (i)%rho (ix,iz))
				vmin = min (vp, vmin)
				vmax = max (vp, vmax)
				if (vs > min_vs) then
					vmin = min (vs, vmin)
					vmax = max (vs, vmax)
				end if
			end do
		end do

		wavel = vmin / (m_s%fpeak*2.5)	! the 2.5 is because of the ricker source. It tapers down at about 2.5 times the central frequency.
		dmax = max (maxval (d_s(i)%dx), maxval (d_s(i)%dz))

    	call from_param("nwavel_samples", nwavel_samples, 8)
	
		if (nwavel_samples*dmax > wavel) then	! at least nwavel_samples per shortest wavelength in field
			write (*,*) 'Reduce dxmax,dymax or dzmax to less than', wavel/nwavel_samples
			call seperr("Grid size too large for accuracy")
		endif
		dmin = min (minval (d_s(i)%dx), minval(d_s(i)%dz))
		if (m_s%dt .gt. 0.4949 * dmin/vmax) then
			write (*,*) 'Reduce dt to less than', 0.4949 * dmin/vmax
			call seperr ('Time Step too large for stability')
		endif
	
		nmax = max (d_s(i)%nz, d_s(i)%nx)
		allocate (d_s(i)%buf1 (nmax*nmax*2))
		allocate (d_s(i)%buf2 (nmax*nmax*2))
		allocate (d_s(i)%all (d_s(i)%nx, d_s(i)%nz, 14))

		d_s(i)%P	=> d_s (i)%all (:,:,P_i)
		d_s(i)%S	=> d_s (i)%all (:,:,S_i)
		d_s(i)%ux	=> d_s (i)%all (:,:,ux_i)
		d_s(i)%uz	=> d_s (i)%all (:,:,uz_i)
		d_s(i)%vx	=> d_s (i)%all (:,:,vx_i)
		d_s(i)%vz	=> d_s (i)%all (:,:,vz_i)
		d_s(i)%txx	=> d_s (i)%all (:,:,txx_i)
		d_s(i)%tzz	=> d_s (i)%all (:,:,tzz_i)
		d_s(i)%txz	=> d_s (i)%all (:,:,txz_i)
		d_s(i)%Upx	=> d_s (i)%all (:,:,Upx_i)
		d_s(i)%Upz	=> d_s (i)%all (:,:,Upz_i)
		d_s(i)%Usx	=> d_s (i)%all (:,:,Usx_i)
		d_s(i)%Usz	=> d_s (i)%all (:,:,Usz_i)
		d_s(i)%uzdz	=> d_s (i)%all (:,:,uzdz_i)


		d_s(i)%all = 0.
	end do

end subroutine


subroutine read_distribute_lam_omp (tag, m_s, d_s)
	character (len=*)   :: tag
	type (model_type)   :: m_s
	type (distrib_type) :: d_s(:)
	integer,external    :: sreed
	integer             :: bz, ez, bx, ex, ith, ix, iz


	if (4 * size (slice2d) /= sreed (tag, slice2d, size (slice2d) * 4)) then
		write (0,*), "=slice tag=", trim(tag)
		call seperr("trouble reading in tag")
	end if

	call expand_slice (slice2d, slice2d_s)

	do ith = 1, size(d_s)
		bz = d_s (ith)%fz + 1
		ez = d_s (ith)%fz + d_s (ith)%nz
		bx = d_s (ith)%fx + 1
		ex = d_s (ith)%fx + d_s (ith)%nx

		d_s (ith)%lambda (:, :) = slice2d_s (bx:ex, bz:ez)
	end do

end subroutine 
subroutine read_distribute_mu_omp (tag, m_s, d_s)
	character (len=*)   :: tag
	type (model_type)   :: m_s
	type (distrib_type) :: d_s(:)
	integer,external    :: sreed
	integer             :: bz, ez, bx, ex, ith, ix, iz


	if (4 * size (slice2d) /= sreed (tag, slice2d, size (slice2d) * 4)) then
		write (0,*), "=slice tag=", trim(tag)
		call seperr("trouble reading in tag")
	end if

	call expand_slice (slice2d, slice2d_s)

	do ith = 1, size(d_s)
		bz = d_s (ith)%fz + 1
		ez = d_s (ith)%fz + d_s (ith)%nz
		bx = d_s (ith)%fx + 1
		ex = d_s (ith)%fx + d_s (ith)%nx

		d_s (ith)%mu (:, :) = slice2d_s (bx:ex, bz:ez)
	end do

end subroutine 
subroutine read_distribute_rho_omp (tag, m_s, d_s)
	character (len=*)   :: tag
	type (model_type)   :: m_s
	type (distrib_type) :: d_s(:)
	integer,external    :: sreed
	integer             :: bz, ez, bx, ex, ith, ix, iz


	if (4 * size (slice2d) /= sreed (tag, slice2d, size (slice2d) * 4)) then
		write (0,*), "=slice tag=", trim(tag)
		call seperr("trouble reading in tag")
	end if

	call expand_slice (slice2d, slice2d_s)

	do ith = 1, size(d_s)
		bz = d_s (ith)%fz + 1
		ez = d_s (ith)%fz + d_s (ith)%nz
		bx = d_s (ith)%fx + 1
		ex = d_s (ith)%fx + d_s (ith)%nx

		d_s (ith)%rho (:, :) = slice2d_s (bx:ex, bz:ez)
	end do

end subroutine 

subroutine expand_slice(small,big)
	real :: small (:,:), big(:,:)
	integer :: ix, iz, nsx, nsz, iloc

	nsx = (size (big, 1) - size (small, 1)) / 2
	nsz = (size (big, 2) - size (small, 2)) / 2

	do iz = 1, size (small, 2)
		do ix = 1, size (small, 1)
			big (ix+nsx, iz+nsz) = small (ix, iz)
		end do
	end do

	do iz = 1, size (big, 2)
		if (iz <= nsz) then
			iloc = nsz+1

		else if (iz > size (big,2) - nsz) then
				iloc = size (small,2)
		else
			iloc = iz - nsz
		end if

		do ix = 1, nsx
			! surf_type = 1 (free) or 0 (absorbing)
			if (surf_type .eq. 1) then
				big (ix, iz) = 0.
			else
				big (ix, iz) = small (1, iloc)
			endif
			big (size (big, 1) - ix+1, iz) = small (size (small, 1), iloc)
		end do

	end do

	do iz = 1, nsz
		do ix = 1, size (big, 1)
			big (ix, iz) = big (ix, nsz+1)
			big (ix, size (big, 2) - iz + 1) = big (ix, size (big, 2) - nsz)
		end do
	end do
end subroutine
end module
