module elas2d_model_mod
	use elas2d_model_type_mod
	use elas2d_distrib_mod
	use elas2d_distrib_type_mod
	use elas2d_thread_type_mod
	use elas2d_io_mod
	use sep
	use sep3d_struct_mod
	use sep_timers_mod
	use helutil_mod
	use opolydiv_mod

	implicit none

	integer							:: tm_u4, tm_v4, tm_sp, tm_ve, tm_tau, tm_se, tm_space, tm_write
	integer, external, private		:: omp_get_thread_num, omp_get_num_threads
	integer, private, save			:: nsx, nsz
	real, pointer, private, save	:: sponge_x(:), sponge_z(:)
	integer, private, save			:: one_step_nthreads
	real							:: decon_norm, eps
	integer							:: ncoeffs, model_jsnap, model_jrec
	type (filter)					:: dfilt
	type (filter)					:: dfilt3d

contains

subroutine model_init(model_s,tag)
	type(model_type)	:: model_s, v_s
	integer, external	:: auxin, sreed
	integer				:: n, irec
	real				:: d
	type(sep3d)			:: mod_s
	real,allocatable	:: junk(:)
	character(len=*)	:: tag
	real, allocatable, dimension (:)	:: tmp1, tmp2, crc, fcoeff
	integer				:: stat, tmpsize
	real				:: ax_0, ax_1, ax_2, az_0, az_1, az_2
	real				:: expnorm


	call init_sep3d (tag, mod_s, "INPUT")
	model_s%nx = mod_s%n(1)
	model_s%nz = mod_s%n(2)


	allocate (model_s%dx (model_s%nx))
	allocate (model_s%dz (model_s%nz))

	if (-1 == auxin ("dx")) then
		model_s%dx = mod_s%d(1)
	else
		call from_aux ("dx", "n1", n)

		if (n /= model_s%nx) &
			call seperr ("number of samples in x not equal n1 of dx")

		if (4 * size (model_s%dx) /= sreed ("dx", model_s%dx, 4 * size (model_s%dx))) &
			call seperr("trouble reading in dx")
	end if

	if (-1 == auxin ("dz")) then
		model_s%dz = mod_s%d(2)
	else
		call from_aux ("dz", "n1", n)

		if (n /= model_s%nz) &
			call seperr ("number of samples in z not equal n1 of dz")
		if (4 * size (model_s%dz) /= sreed ("dz", model_s%dz, 4 * size (model_s%dz))) &
			call seperr ("trouble reading in dz")
	end if

	call mk_sponge(model_s)

	!expand ds
	allocate(junk(size(model_s%dz)))
	junk=model_s%dz
	deallocate(model_s%dz)

	model_s%nz=model_s%nz+size(sponge_z)*2
	allocate(model_s%dz(model_s%nz))
	model_s%dz(model_s%nsponge_z+1:model_s%nz-model_s%nsponge_z)=junk
	model_s%dz(1:model_s%nsponge_z)=junk(1)
	model_s%dz(model_s%nz-model_s%nsponge_z+1:)=junk(size(junk))
	deallocate(junk)

	allocate(junk(size(model_s%dx)))
	junk=model_s%dx
	deallocate(model_s%dx)

	model_s%nx=model_s%nx+size(sponge_x)*2
	allocate(model_s%dx(model_s%nx))
	model_s%dx(model_s%nsponge_x+1:model_s%nx-model_s%nsponge_x)=junk
	model_s%dx(1:model_s%nsponge_x)=junk(1)
	model_s%dx(model_s%nx-model_s%nsponge_x+1:)=junk(size(junk))
	deallocate(junk)

	call from_param("nt",model_s%nt)
	call from_param("dt",model_s%dt)

	call from_param("fpeak",model_s%fpeak,50.)
	call make_wavelet(model_s)

	call from_aux("recs","n1",n)
	if(n/=2) call seperr("expecting receiver file to have n1 = 2")

	call from_aux("recs","n2",n)
	allocate(model_s%recs(2,n))

	if (4 * size (model_s%recs) /= sreed ("recs", model_s%recs, 4 * size (model_s%recs))) &
		call seperr("trouble reading in recs")

	do irec = 1, n
		model_s%recs(1,irec)=model_s%recs(1,irec)+model_s%nsponge_x
		model_s%recs(2,irec)=model_s%recs(2,irec)+model_s%nsponge_z
	end do

	call from_param("isrc_type",model_s%isrc_type,0)

	model_s%ixsrc=model_s%nx/2+model_s%nsponge_x
	model_s%izsrc=model_s%nz/2+model_s%nsponge_z

	call from_param("ixsrc",model_s%ixsrc,model_s%ixsrc)
	call from_param("izsrc",model_s%izsrc,model_s%izsrc)

	model_s%ixsrc=model_s%ixsrc+model_s%nsponge_x
	model_s%izsrc=model_s%izsrc+model_s%nsponge_z

	allocate(model_s%rec_i(n))
	allocate(model_s%rec_ith(n))

	call from_param ("proptype", model_s%proptype, source)

	call from_param ("one_step_nthreads", one_step_nthreads, 1)
	call omp_set_num_threads (one_step_nthreads)

	call from_param ("jsnap",model_jsnap)
	call from_param ("jrec",model_jrec)

	! ======================================
	! decon filter construction
	! ======================================
	call from_param ("splitdisp", model_s%splitdisp, 0)
	call from_param ("ncoeffs", ncoeffs, 5)
	call from_param ("eps", eps, 1e-20)

	ax_2 = -1. / (model_s%dx(1) * model_s%dx(1) * 12.)
	ax_1 = -16. * ax_2
	ax_0 = 30. * ax_2
	az_2 = -1. / (model_s%dz(1) * model_s%dz(1) * 12.)
	az_1 = -16. * az_2
	az_0 = 30. * az_2

	! make 4th order filter
	dfilt = lapfac4 (eps, ax_0 + az_0, ax_1, az_1, ax_2, az_2, model_s%nx, ncoeffs)

	allocate (fcoeff (size(dfilt%flt) + 1))
	allocate (crc    (size (fcoeff)))
	
	fcoeff (1) = 1.
	fcoeff (2 : size (dfilt%flt) + 1) = dfilt%flt
	
	call correlate (fcoeff, fcoeff, crc, size(fcoeff))
	expnorm = (az_0 + ax_0 + eps) / crc(1)
	crc = crc * expnorm
	!write (0,*), "filter autocorrelation after norm = "
	!write (0,*), crc
	deallocate (crc)
	deallocate (fcoeff)

	! the decon result needs to be normalized, since the zero-lag cross-correlation element is an implied 1 in the polydiv code.
	! The normalization changes it so that the it comes out the same as if the zero-lag cross-correlation element was the central finite difference weight.

	tmpsize = ncoeffs * 2 + 1 ! an odd number, so that we'll have a central array element.
	call opolydiv_init (tmpsize, dfilt)

	allocate (tmp1 (tmpsize))
	allocate (tmp2 (tmpsize))
	tmp1 = .0
	tmp2 = .0

	tmp1 (tmpsize / 2 + 1) = 1.

	! run the decon on a spike
	stat = opolydiv_lop (.true.,  .false., tmp2, tmp1)
	stat = opolydiv_lop (.false., .false., tmp2, tmp1)

	! The normalization factor is the ratio between the deconvolved spike and the central FD weight.
	decon_norm = tmp1 (tmpsize / 2 + 1)  / (az_0 + ax_0 + eps)

	deallocate (tmp1)
	deallocate (tmp2)

	call opolydiv_close ()
end subroutine

subroutine make_wavelet (model_s)
	type(model_type)	:: model_s
	integer				:: ntw, i
	real				:: a, b, t2

	ntw = 1. / model_s%fpeak / model_s%dt
	allocate (model_s%wavelet (ntw*2 + 1))
	b = (3.1415926 * model_s%fpeak)**2
	a = 2. * b
	model_s%wavelet = 0
	do  i = 1, ntw+1
		t2 = ((i-1) * model_s%dt)**2
		model_s%wavelet (ntw+i)   = (1. - a * t2) * exp(-b * t2)
		model_s%wavelet (ntw+2-i) = model_s%wavelet (ntw+i)
	end do

end subroutine

! =============================================================
!     vg_coeffz
!     calculate coeff. of 4th order FD operator for varying grid (z)
!     Pitarka, BSSA, pp54-68, 1999
! =============================================================
 subroutine fdvsg3d_vg_coeffz(myd)
  real c(4)
  integer  :: i,j
  real     :: d1,d2,d3,d4
  type(distrib_type) :: myd

  allocate(myd%cz1(myd%nz,4))
  allocate(myd%cz2(myd%nz,4))

  !...  array cz1(,) is coeff. at the middle of grids
  do i=2,myd%nz-2
    d1=myd%dz(i+1)+myd%dz(i)/2.
    d2=myd%dz(i-1)+myd%dz(i)/2.
    d3=myd%dz(i)/2.
    d4=myd%dz(i)/2.
    call fdvsg3d_get_coeff(d1,d2,d3,d4,c)
    myd%cz1(i,:)=c(:)
  end do

  !array cz2(,) is coeff. at grid
  do i=3,myd%nz-2
    d1=myd%dz(i)+myd%dz(i+1)/2.
    d2=myd%dz(i-1)+myd%dz(i-2)/2.
    d3=myd%dz(i)/2.
    d4=myd%dz(i-1)/2.
    call fdvsg3d_get_coeff(d1,d2,d3,d4,c)
    myd%cz2(i,:)=c
 enddo
end subroutine

! =============================================================
!     vg_coeffx
!     calculate coeff. of 4th order FD operator for varying grid (x)
!     Pitarka, BSSA, pp54-68, 1999
! =============================================================
 subroutine fdvsg3d_vg_coeffx(myd)
  real c(4)
  integer  :: i,j
  real     :: d1,d2,d3,d4
  type(distrib_type) :: myd

  allocate(myd%cx1(myd%nx,4))
  allocate(myd%cx2(myd%nx,4))

  !...  array cx1(,) is coeff. at the middle of grids
  do i=2,myd%nx-2
    d1=myd%dx(i+1)+myd%dx(i)/2.
    d2=myd%dx(i-1)+myd%dx(i)/2.
    d3=myd%dx(i)/2.
    d4=myd%dx(i)/2.
    call fdvsg3d_get_coeff(d1,d2,d3,d4,c)
    myd%cx1(i,:)=c(:)
  end do

  !array cx2(,) is coeff. at grid
  do i=3,myd%nx-2
    d1=myd%dx(i)+myd%dx(i+1)/2.
    d2=myd%dx(i-1)+myd%dx(i-2)/2.
    d3=myd%dx(i)/2.
    d4=myd%dx(i-1)/2.
    call fdvsg3d_get_coeff(d1,d2,d3,d4,c)
    myd%cx2(i,:)=c
 enddo

end subroutine

! =============================================================
!     get_coeff
!     solve linear equations to get coeff. cx1(,)/cx2(,)
! =============================================================

subroutine fdvsg3d_get_coeff(d1,d2,d3,d4,c)
	real c(4),d1,d2,d3,d4

	c(1) = (d3 * (d4 + d2) - d4 * d2) / (d3 - d1) / (d2 + d1) / (d4 + d1)
	c(3) = (d4 * d2 - d1 * (d4 + d2)) / (d3 - d1) / (d4 + d3) / (d3 + d2)
	c(4) = (1. - c(3) * (d2 + d3) - (d2 + d1) * c(1)) / (d2 - d4)
	c(2) = -c(1) - c(4) - c(3)

end subroutine

subroutine one_step (ith, it, model_s, distrib_s, io_s)
	integer						:: ith, it, thrthr
	type(model_type)			:: model_s
	type(distrib_type),pointer	:: distrib_s(:), myd
	type(io_type)				:: io_s
	real						:: Psrc, Vsrc


	myd=>distrib_s(ith)

	!add dipole source / geophone data into wavefield
	if (model_s%proptype == source) then
		if (myd%ixsrc >0       .and. myd%izsrc>0         .and. &
			myd%ixsrc <=myd%nx .and. myd%izsrc<=myd%nz ) then

			Vsrc = model_s%wavelet(it) / (myd%lambda (myd%ixsrc, myd%izsrc) + 2 * myd%mu (myd%ixsrc, myd%izsrc)) / myd%dx (myd%ixsrc) / myd%dz (myd%izsrc)
			!Vsrc = model_s%wavelet(it) * myd%dt * myd%rho (myd%ixsrc, myd%izsrc) / myd%dx (myd%ixsrc) / myd%dz (myd%izsrc)
			!Vsrc = model_s%wavelet(it) / myd%dx (myd%ixsrc) / myd%dz (myd%izsrc)
			if (model_s%isrc_type .eq. 1 .and. it .le. size(model_s%wavelet)) then		! source in x direction

				myd%vx (myd%ixsrc, myd%izsrc) = myd%vx (myd%ixsrc, myd%izsrc) + 0.02 * Vsrc
	
			else if  (model_s%isrc_type .eq. 3 .and. it .le. size (model_s%wavelet)) then	! source in z direction

				myd%vz (myd%ixsrc, myd%izsrc) = myd%vz (myd%ixsrc, myd%izsrc) + 0.02 * Vsrc
			endif
		endif
	else

		!add geophone trace data. The trace data must be interpolated to be in modeling dt intervals, otherwise we'll reach EOF.
		call read_recs ("in_vx", io_s%in_vx, ith, model_s, distrib_s, vx_i, -1)
		call read_recs ("in_vz", io_s%in_vz, ith, model_s, distrib_s, vz_i, -1)
	endif

	!advance velocities of velocity field
	if(omp_get_thread_num()==0) call start_timer_num(tm_v4)
	call fdvsg3d_v4 (myd%nz, myd%nx, myd%rho, myd%cx1, myd%cx2, myd%cz1, myd%cz2, myd%ux, myd%uz, myd%vx, myd%vz, myd%txx, myd%tzz, myd%txz, myd%dt)
	if(omp_get_thread_num()==0) call stop_timer_num(tm_v4)

	!do velocity boundary conditions on velocity field
	if(omp_get_thread_num()==0) call start_timer_num(tm_sp)
	call fdvsg3d_sponge (myd,myd%vx)
	call fdvsg3d_sponge (myd,myd%vz)
	if(omp_get_thread_num()==0) call stop_timer_num(tm_sp)

	if (model_s%splitdisp .eq. 1 .and. (io_s%jrec == 1 .or. mod (it, io_s%jrec) == 1)) then
	call fdvsg3d_disp_space (myd%nz, myd%nx, myd%cx1, myd%cx2, myd%cz1, myd%cz2, myd%ux, myd%uz, myd%Upx, myd%Upz, myd%Usx, myd%Usz)
	endif

	!add pressure source / hydrophone data into stress field
	if (model_s%proptype == source) then
		if (myd%ixsrc > 0       .and. myd%izsrc > 0         .and. &
			myd%ixsrc <= myd%nx .and. myd%izsrc <= myd%nz ) then
	
			if (model_s%isrc_type .eq. 0 .and. it .le. size (model_s%wavelet)) then

				!Psrc = model_s%wavelet (it) * (myd%lambda (myd%ixsrc, myd%izsrc) + 2 * myd%mu (myd%ixsrc, myd%izsrc)) * &
				!		myd%rho (myd%ixsrc, myd%izsrc) / model_s%dz (myd%izsrc) / myd%dx (myd%ixsrc)
				Psrc = model_s%wavelet (it) * model_s%dt / myd%dz (myd%izsrc) / myd%dx (myd%ixsrc)

				myd%txx (myd%ixsrc, myd%izsrc) = myd%txx (myd%ixsrc, myd%izsrc) + Psrc
	
				myd%tzz (myd%ixsrc, myd%izsrc) = myd%tzz (myd%ixsrc, myd%izsrc) + Psrc
			endif
		endif
	else

		!add hydrophone trace data. The trace data must be interpolated to be in modeling dt intervals, otherwise we'll reach EOF.
		call read_recs ("in_P", io_s%in_P, ith, model_s, distrib_s, txx_i, tzz_i)
	endif

	!advance stresses of stress field
	if(omp_get_thread_num()==0) call start_timer_num(tm_tau)
	call fdvsg3d_tau4 (myd%nz, myd%nx, myd%lambda, myd%mu, myd%cx1, myd%cx2, myd%cz1, myd%cz2, myd%ux, myd%uz, myd%vx, myd%vz, &
 					   myd%txx, myd%tzz, myd%txz, myd%dt, myd%P, myd%S, myd%uzdz)
	if(omp_get_thread_num()==0) call stop_timer_num(tm_tau)

	!do stress boundary conditions on stress field
	if(omp_get_thread_num()==0) call start_timer_num(tm_sp)
	call fdvsg3d_sponge(myd,myd%txx)
	call fdvsg3d_sponge(myd,myd%tzz)
	call fdvsg3d_sponge(myd,myd%txz)
	if(omp_get_thread_num()==0) call stop_timer_num(tm_sp)

	return
end subroutine


! =============================================================
!     v4
!     update vx, vz, vy using 4th order
! =============================================================

subroutine fdvsg3d_v4 (nz, nx, rho, cx1, cx2, cz1, cz2, ux, uz, vx, vz, txx, tzz, txz, dt)
	real, parameter			:: c0=9./8., c1=1./24.
	integer					:: i, k, nx, nz
	real					:: dt, bx, bz, dx_txx, dz_txz, dx_txz, dz_tzz
	real, dimension (:,:)	:: cx1, cx2, cz1, cz2
	real, dimension (:,:)	:: rho, ux, uz, vx, vz, txx, tzz, txz

!$OMP PARALLEL private (k, i, bx, dx_txx, dz_txz)
!$OMP DO
	do k=3,nz-2
		do i=3,nx-3
			bx     = 0.5 * (rho (i+1, k) + rho (i, k))
			dx_txx = cx1 (i,1) * txx (i+2, k)   + cx1 (i,2) * txx (i-1, k)   + &
					 cx1 (i,3) * txx (i+1, k)   + cx1 (i,4) * txx (i,   k)
			dz_txz = cz2 (k,1) * txz (i,   k+1) + cz2 (k,2) * txz (i,   k-2) + &
					 cz2 (k,3) * txz (i,   k)   + cz2 (k,4) * txz (i,   k-1)

			vx (i,k) = vx (i,k) + dt * bx * (dx_txx + dz_txz) !update velocity by integrating acceleration
			ux (i,k) = ux (i,k) + dt * vx(i,k)
		end do
	end do
!$OMP END DO
!$OMP END PARALLEL


!$OMP PARALLEL private (k, i, bz, dx_txz, dz_tzz)
!$OMP DO
	do k=3,nz-3
		do i=3,nx-2
			bz     = 0.5 * (rho (i, k) + rho (i, k+1))
			dx_txz = cx2 (i,1) * txz (i+1, k)   + cx2 (i,2) * txz (i-2, k)   + &
					 cx2 (i,3) * txz (i,   k)   + cx2 (i,4) * txz (i-1, k)
			dz_tzz = cz1 (k,1) * tzz (i,   k+2) + cz1 (k,2) * tzz (i,   k-1) + &
					 cz1 (k,3) * tzz (i,   k+1) + cz1 (k,4) * tzz (i,   k)

			vz (i,k) = vz (i,k) + dt * bz * (dx_txz + dz_tzz) !update velocity by integrating acceleration
			uz (i,k) = uz (i,k) + dt * vz(i,k)
		end do
	end do
!$OMP END DO
!$OMP END PARALLEL
end subroutine

! ===============================================================
!     tau4
!     update txx, tzz, txz using 4th order
!     or add stress perturbation as source function to dtxx, dtzz, dtxz using 4th order
! ===============================================================
subroutine fdvsg3d_tau4 (nz, nx, lambda, mu, cx1, cx2, cz1, cz2, ux, uz, vx, vz, txx, tzz, txz, dt, P, S, uzdz)
	real,  parameter		:: c0 = 9./8., c1=1./24.
	integer					:: i, k, nz, nx
	real					:: dt, r2u, rta
	real					:: u14, u, Vx_dx, Vz_dz, Vx_dz, Vz_dx, Ux_dx, Uz_dz, Ux_dz, Uz_dx
	real, dimension (:,:)	:: cx1, cx2, cz1, cz2
	real, dimension (:,:)	:: lambda, mu, ux, uz, vx, vz, txx, tzz, txz, P, S, uzdz

	! calculate normal stresses txx, tzz
!$OMP PARALLEL private (k, i, r2u, rta, Vx_dx, Vz_dz)
!$OMP DO
	do k = 3, nz-2
		do i = 3, nx-2
			r2u = lambda (i, k) + 2 * mu (i, k)
			rta = lambda (i, k)

			Vx_dx = cx2 (i, 1) * vx (i+1, k)   + cx2 (i, 2) * vx (i-2, k)   + &
					cx2 (i, 3) * vx (i,   k)   + cx2 (i, 4) * vx (i-1, k)
			Vz_dz = cz2 (k, 1) * vz (i,   k+1) + cz2 (k, 2) * vz (i,   k-2) + &
					cz2 (k, 3) * vz (i,   k)   + cz2 (k, 4) * vz (i,   k-1)

			txx (i, k) = txx (i, k) + dt * (r2u * Vx_dx + rta * Vz_dz)
			tzz (i, k) = tzz (i, k) + dt * (r2u * Vz_dz + rta * Vx_dx)
			!P   (i, k) = 0.5 * (txx (i, k) + tzz (i, k))

			Ux_dx = cx2 (i, 1) * ux (i+1, k)   + cx2 (i, 2) * ux (i-2, k)   + &
					cx2 (i, 3) * ux (i,   k)   + cx2 (i, 4) * ux (i-1, k)
			Uz_dz = cz2 (k, 1) * uz (i,   k+1) + cz2 (k, 2) * uz (i,   k-2) + &
					cz2 (k, 3) * uz (i,   k)   + cz2 (k, 4) * uz (i,   k-1)
			!P   (i, k) = 0.5 * (r2u * (Ux_dx + Uz_dz) + rta * (Uz_dz + Ux_dx))
			P   (i, k) = 0.5 * ((Ux_dx + Uz_dz) + (Uz_dz + Ux_dx)) ! this is dimensionless P-wave amplitude
			uzdz (i, k) = Uz_dz ! horizontal derivative of displacement


		end do
	end do
!$OMP END DO
!$OMP END PARALLEL

	! calculate shear stress txz

!$OMP PARALLEL private  (k, i, u14, u, Vx_dz, Vz_dx)
!$OMP DO
	do k = 2, nz-2
		do i = 2, nx-2
			u14 = mu (i, k) * mu (i+1, k) * mu (i, k+1) * mu (i+1, k+1)

			if (abs (sum (mu (i:i+1, k:k+1)) - 4 * mu (i, k)) <.0001) then ! if this mu is near in value to its neighbours, use this mu
				u = mu (i, k)
			else
				if (u14 .le. 1.e-4) then ! if this mu is different in value from the neighbouring mus, but the neighbourhood has small values, then use mu=0
					u = 0.
				else ! otherwise, use harmonic average for this mu
					u = (1. / mu (i, k) + 1. / mu (i+1, k) + 1. / mu (i, k+1) + &
						1. / mu (i+1, k+1)) / 4.
					u = 1. / u
				endif
			end if

			Vx_dz = cz1 (k, 1) * vx (i,   k+2) + cz1 (k, 2) * vx (i,   k-1) + &
					cz1 (k, 3) * vx (i,   k+1) + cz1 (k, 4) * vx (i,   k)
			Vz_dx = cx1 (i, 1) * vz (i+2, k)   + cx1 (i, 2) * vz (i-1, k)   + &
					cx1 (i, 3) * vz (i+1, k)   + cx1 (i, 4) * vz (i,   k)

			txz (i, k) = txz (i, k) + dt * u * (Vz_dx + Vx_dz)
			!S (i, k) = dt * u * (Vx_dz - Vz_dx)

			Ux_dz = cz1 (k, 1) * ux (i,   k+2) + cz1 (k, 2) * ux (i,   k-1) + &
					cz1 (k, 3) * ux (i,   k+1) + cz1 (k, 4) * ux (i,   k)
			Uz_dx = cx1 (i, 1) * uz (i+2, k)   + cx1 (i, 2) * uz (i-1, k)   + &
					cx1 (i, 3) * uz (i+1, k)   + cx1 (i, 4) * uz (i,   k)
			!S   (i, k) = u * (Ux_dz - Uz_dx)

			if (u .gt. 1.e-4) then
				S (i, k) = (Ux_dz - Uz_dx) ! this is dimensionless shear wave amplitude
			endif

		end do
	end do
!$OMP END DO
!$OMP END PARALLEL

end subroutine

! =============================================================
!	disp_space
!	do displacement decomposition in space domain.
! =============================================================

subroutine fdvsg3d_disp_space (nz, nx, cx1, cx2, cz1, cz2, vx, vz, Vpx, Vpz, Vsx, Vsz)
	integer					:: nz, nx
	real, dimension (:,:)	:: cx1, cx2, cz1, cz2
	real, dimension (:,:)	:: vx, vz, Vpx, Vpz, Vsx, Vsz
	real, save, allocatable, dimension (:)	:: vx_1d, vz_1d, tmp
	real, save, allocatable, dimension (:,:):: Vx_dx, Vz_dz, Vx_dz, Vz_dx
	integer		:: i,k
	integer		:: stat
	real		:: Vx_d2x, Vz_d2z, Vx_d2z, Vz_d2x, Vz_dzx, Vx_dxz

	if (.not. allocated (vx_1d)) then
		allocate (vx_1d (nx * nz))
		allocate (vz_1d (nx * nz))
		allocate (Vx_dx (nx, nz))
		allocate (Vz_dz (nx, nz))
		allocate (Vx_dz (nx, nz))
		allocate (Vz_dx (nx, nz))
		allocate (tmp    (nx * nz))
		call opolydiv_init (nx * nz, dfilt)
		!write (0,*), "filter%flt =", dfilt%flt
	endif

	! integrate vx and vz to get vx and vz
!$OMP PARALLEL private (k, i)
!$OMP DO
	do k=1,nz
		do i=1,nx
			vx_1d ((k-1)*nx + i) = vx (i, k) ! put into 1d array for helical decon
			vz_1d ((k-1)*nx + i) = vz (i, k)
		end do
	end do
!$OMP END DO
!$OMP END PARALLEL

	! apply helical deconvolution of 4th order spatial derivative operator to vx and vz - this is equivalent to the normalization by |K| in wavenumber domain
	! For now, this is done only to one Y-plane of the 3D model ( j = iysrc ).  (wax-off, wax-on)
 	tmp = 0.
 	stat = opolydiv_lop (.true., .false.,  tmp, vx_1d) ! yes, we can certainly save here by reusing vz_1d. I'll do it later when I get the hang of things (promise)
 	stat = opolydiv_lop (.false., .false., tmp, vx_1d)
 	tmp = 0.
 	stat = opolydiv_lop (.true., .false.,  tmp, vz_1d)
 	stat = opolydiv_lop (.false., .false., tmp, vz_1d)

	! first derivative d(Ux)/dx and d(Uz)/dz. Their sum is used for P wave amplitude, and later for the P-S displacement decomposition
	! first derivative d(Ux)/dz and d(Uz)/dx. Their diff is used for S wave amplitude, and later for the P-S displacement decomposition
!$OMP PARALLEL private (k, i)
!$OMP DO
	do k = 3, nz-2
		do i = 3, nx-2

			! first derivative of deconvolved displacement field for P-S decomposition
			Vx_dx (i, k) = (cx2 (i, 1) * vx_1d ((k-1)*nx + i+1) + cx2 (i, 2) * vx_1d ((k-1)*nx + i-2)   + &
							cx2 (i, 3) * vx_1d ((k-1)*nx + i)   + cx2 (i, 4) * vx_1d ((k-1)*nx + i-1)) * decon_norm
			
			Vz_dz (i, k) = (cz2 (k, 1) * vz_1d ((k  )*nx + i) + cz2 (k, 2) * vz_1d ((k-3)*nx + i) + &
							cz2 (k, 3) * vz_1d ((k-1)*nx + i) + cz2 (k, 4) * vz_1d ((k-2)*nx + i)) * decon_norm


			! first derivative of deconvolved displacement field for P-S decomposition
			Vx_dz (i, k) = (cz1 (k, 1) * vx_1d ((k+1)*nx + i) + cz1 (k, 2) * vx_1d ((k-2)*nx + i) + &
							cz1 (k, 3) * vx_1d ((k  )*nx + i) + cz1 (k, 4) * vx_1d ((k-1)*nx + i)) * decon_norm
			
			Vz_dx (i, k) = (cx1 (i, 1) * vz_1d ((k-1)*nx + i+2) + cx1 (i, 2) * vz_1d ((k-1)*nx + i-1)   + &
							cx1 (i, 3) * vz_1d ((k-1)*nx + i+1) + cx1 (i, 4) * vz_1d ((k-1)*nx + i  )) * decon_norm

		end do
	end do
!$OMP END DO
!$OMP END PARALLEL

	! second derivative d^2(Ux)/dx^2 at vx grid points, and d^2(Ux)/dz^2 at vx grid points
	! second derivative d^2(Uz)/dz^2 at vx grid points, and d^2(Uz)/dx^2 at vz grid points

!$OMP PARALLEL private (k, i, Vx_d2x, Vx_d2z, Vx_dxz, Vz_d2z, Vz_d2x, Vz_dzx)
!$OMP DO
	do k = 3, nz-2
		do i = 3, nx-2

			Vx_d2x = cx1 (i, 1) * Vx_dx (i+2, k)   + cx1 (i, 2) * Vx_dx (i-1, k)	+ &
				     cx1 (i, 3) * Vx_dx (i+1, k)   + cx1 (i, 4) * Vx_dx (i,   k)

			Vx_d2z = cz2 (k, 1) * Vx_dz (i,   k+1) + cz2 (k, 2) * Vx_dz (i,   k-2)	+ &
				     cz2 (k, 3) * Vx_dz (i,   k)   + cz2 (k, 4) * Vx_dz (i,   k-1)

			Vx_dxz = cz1 (k, 1) * Vx_dx (i,   k+2) + cz1 (k, 2) * Vx_dx (i,   k-1)	+ &
				     cz1 (k, 3) * Vx_dx (i,   k+1) + cz1 (k, 4) * Vx_dx (i,   k)

			Vz_d2z = cz1 (k, 1) * Vz_dz (i,   k+2) + cz1 (k, 2) * Vz_dz (i,   k-1)	+ &
				     cz1 (k, 3) * Vz_dz (i,   k+1) + cz1 (k, 4) * Vz_dz (i,   k)

			Vz_d2x = cx2 (i, 1) * Vz_dx (i+1, k)   + cx2 (i, 2) * Vz_dx (i-2, k)	+ &
				     cx2 (i, 3) * Vz_dx (i,   k)   + cx2 (i, 4) * Vz_dx (i-1, k)

			Vz_dzx = cx1 (i, 1) * Vz_dz (i+2, k)   + cx1 (i, 2) * Vz_dz (i-1, k)	+ &
				     cx1 (i, 3) * Vz_dz (i+1, k)   + cx1 (i, 4) * Vz_dz (i,   k)

			! displacement separation (Zhang and McMechan, 2010)
			Vpx (i, k) = Vx_d2x + Vz_dzx		
			Vpz (i, k) = Vz_d2z + Vx_dxz
			Vsx (i, k) = Vx_d2z - Vz_dzx
			Vsz (i, k) = Vz_d2x - Vx_dxz
		end do
	end do
!$OMP END DO
!$OMP END PARALLEL

end subroutine

subroutine fdvsg3d_sponge(myd,array)
	type(distrib_type) :: myd
	real :: array(:,:)
	integer  :: ix,iy,iz

	!. Top corrosponds to null proc in z-minus direction
	if (myd%lz .eq. -1) call fdvsg3d_vabs_z_l(myd,array)

	!. Bottom corrosponds to null proc in z-minus direction
	if (myd%hz .eq. -1) call fdvsg3d_vabs_z_u(myd,array)

	!. Top corrosponds to null proc in x-minus direction
	if (myd%lx .eq. -1) call fdvsg3d_vabs_x_l(myd,array)

	!. Bottom corrosponds to null proc in x-minus direction
	if (myd%hx .eq. -1) call fdvsg3d_vabs_x_u(myd,array)
end subroutine

! ======================================================================
!     vabs_x_l
!     dump velocity near top boundary
!     using Cerjan et al. sponge absorbing boundary conditions
!     (see also mksponge)
! ======================================================================

subroutine fdvsg3d_vabs_x_l(myd,array)
	real :: array(:,:)
	type(distrib_type) :: myd
	integer :: ix,iz

!$OMP PARALLEL private (iz, ix)
!$OMP DO
	do iz=1,myd%nz
		do ix=1,nsx
			array (ix,iz)=array(ix,iz)*sponge_x(ix)
		end do
	end do
!$OMP END DO
!$OMP END PARALLEL
end subroutine



! ======================================================================
!     vabs_bo
!     dump velocity near bottom boundary
!     using Cerjan et al. sponge absorbing boundary conditions
!     (see also mksponge)
! ======================================================================
subroutine fdvsg3d_vabs_x_u(myd,array)
	type(distrib_type) :: myd
	real :: array(:,:)
	integer :: iz,ix


!$OMP PARALLEL private (iz, ix)
!$OMP DO
	do iz=1,myd%nz
		do ix=1,nsx
			array(myd%nx+1-ix,iz)=array(myd%nx+1-ix,iz)*sponge_x(ix)
		end do
	end do
!$OMP END DO
!$OMP END PARALLEL
end subroutine

! ======================================================================
!     vabs_z_l
!     dump velocity near top boundary
!     using Cerjan et al. sponge absorbing boundary conditions
!     (see also mksponge)
! ======================================================================

subroutine fdvsg3d_vabs_z_l(myd,array)
	type(distrib_type) :: myd
	integer :: iz,ix
	real :: array(:,:)


!$OMP PARALLEL private (iz, ix)
!$OMP DO
	do iz=1,nsz
		do ix=1+nsx,myd%nx-nsx
			array(ix,iz)=array(ix,iz)*sponge_z(iz)
		end do
	end do
!$OMP END DO
!$OMP END PARALLEL
end subroutine



! ======================================================================
!     vabs_bo
!     dump velocity near bottom boundary
!     using Cerjan et al. sponge absorbing boundary conditions
!     (see also mksponge)
! ======================================================================
subroutine fdvsg3d_vabs_z_u(myd,array)
	type(distrib_type) :: myd
	real :: array(:,:)
	integer :: iz,ix


!$OMP PARALLEL private (iz, ix)
!$OMP DO
	do iz=1,nsz
		do ix=1+nsx,myd%nx-nsx
			array(ix,myd%nz+1-iz)=array(ix,myd%nz+1-iz)*sponge_z(iz)
		end do
	end do
!$OMP END DO
!$OMP END PARALLEL
end subroutine

subroutine mk_sponge(model_s)
	type(model_type) :: model_s
	real  :: xlambda,a0,damper
	integer :: i,imax
	real	:: pi = 4. * atan(1.)

	call from_param("xlambda",xlambda,10.*model_s%dz(1))
	call from_param("damper",damper,0.6)

	imax=2*xlambda/model_s%dz(size(model_s%dz)-1)
	a0=damper/real(imax)

	allocate(sponge_z(imax))
	model_s%nsponge_z=imax
	do i=1,imax
		sponge_z(i)=exp(-(a0*real(imax+1-i))**2)
		!sponge_z(i) = 1. - damper * (1. + cos (pi * i / (imax+1)))
	end do

	imax=2*xlambda/model_s%dx(size(model_s%dx)-1)
	a0=damper/real(imax)

	allocate(sponge_x(imax))
	model_s%nsponge_x=imax
	do i=1,imax
		sponge_x(i)=exp(-(a0*real(imax+1-i))**2)
		!sponge_x(i) = 1. - damper * (1. + cos (pi * i / (imax+1)))
	end do

	nsx=size(sponge_x)
	nsz=size(sponge_z)

end subroutine


subroutine print_it(array,string)
	real :: array(:,:)
	character(len=*) :: string 

	write(0,*) trim(string),sum(array),omp_get_thread_num(),maxloc(array)

	!write(0,*) trim(string),array(120,120,53),size(array,1),size(array,2),size(array,3),omp_get_thread_num(),maxval(array(20:,20:,20:))

end subroutine

integer function pad2 (n)
	integer	:: n
	pad2 = 1

	do while (pad2 .lt. n)
		pad2 = pad2*2
	end do

	return
end function pad2

end module
