module model_mod
	use model_type_mod
	use distrib_type_mod
	use io_mod
	use sep
	use sep3d_struct_mod
	use sep_timers_mod
	use helutil_mod
	use opolydiv_mod
	use adj_mod

	implicit none

	integer							:: tm_u4, tm_v4, tm_sp, tm_ve, tm_se, tm_space, tm_write, tm_read, tm_cgstep
	integer, external, private		:: omp_get_thread_num, omp_get_num_threads
	integer, private, save			:: nrec, nsrc
	integer, private, save			:: one_step_nthreads
	real							:: eps
	type (filter)					:: dfilt
	real, private, save				:: ax1, ax2, az1, az2, axz, apx1, apz1, asx1, asz1
	integer							:: ncoeffs
	type (model_type)				:: m_s
	type (distrib_type)				:: d_s
	type (io_type)					:: io_s

contains

subroutine model_init (m_s_in, tag)
	type (model_type)	:: m_s_in
	integer, external	:: sreed
	integer				:: n, irec, isrc
	real				:: d
	type (sep3d)		:: mod_s
	character (len=*)	:: tag


	call from_param ("adj", m_s_in%adj, .false.)
	call from_param ("add", m_s_in%add, .false.)

	! read dimensions from one of the medium model files
	call init_sep3d (tag, mod_s, "INPUT")
	m_s_in%nx = mod_s%n(1)
	m_s_in%nz = mod_s%n(2)
	m_s_in%dx = mod_s%d(1)
	m_s_in%dz = mod_s%d(2)

	call mk_sponge(m_s_in)

	!expand to include sponge size
	m_s_in%nz = m_s_in%nz + 2 * m_s_in%nsponge_z
	m_s_in%nx = m_s_in%nx + 2 * m_s_in%nsponge_x

	! first index. This is a leftover from the MPI version
	m_s_in%fx = 1
	m_s_in%fz = 1

	call from_param("nt",m_s_in%nt)
	call from_param("dt",m_s_in%dt)
	call from_param("fpeak",m_s_in%fpeak,50.)
	call make_wavelet(m_s_in)

	! point source type. 0=pressure. 1=vertical. 3=horizontal.
	call from_param("isrc_type", m_s_in%isrc_type, 0)

	! point source location. default: at center of unpadded field
	call from_param("ixsrc", m_s_in%ixsrc, m_s_in%nx/2 - m_s_in%nsponge_x)
	call from_param("izsrc", m_s_in%izsrc, m_s_in%nz/2 - m_s_in%nsponge_z)

	! shift source location to padded field coordinates
	m_s_in%ixsrc = m_s_in%ixsrc + m_s_in%nsponge_x
	m_s_in%izsrc = m_s_in%izsrc + m_s_in%nsponge_z

	! input wavelet from point source or from virtual source in the forward modeling operation
	call from_param ("proptype", m_s_in%proptype, source)

	! do P/S displacement decomposition
	call from_param ("splitdisp", m_s_in%splitdisp, 0)

	call from_param ("one_step_nthreads", one_step_nthreads, 1)
	call omp_set_num_threads (one_step_nthreads)

	!##### read receiver locations
	call from_aux("recs","n1",n)
	if(n/=2) call seperr("expecting receiver file to have n1 = 2")

	call from_aux("recs","n2",nrec)
	allocate(m_s_in%recs(2,nrec))

	if (4 * size (m_s_in%recs) /= sreed ("recs", m_s_in%recs, 4 * size (m_s_in%recs))) &
		call seperr("model_mod: trouble reading in recs")

	do irec = 1, nrec
		m_s_in%recs(1,irec)=m_s_in%recs(1,irec)+m_s_in%nsponge_x
		m_s_in%recs(2,irec)=m_s_in%recs(2,irec)+m_s_in%nsponge_z
	end do

	!##### read virtual source locations
	call from_aux("srcs","n1",n)
	if(n/=2) call seperr("expecting receiver file to have n1 = 2")

	call from_aux("srcs","n2",nsrc)
	allocate(m_s_in%srcs(2,nsrc))

	if (4 * size (m_s_in%srcs) /= sreed ("srcs", m_s_in%srcs, 4 * size (m_s_in%srcs))) &
		call seperr("model_mod: trouble reading in srcs")

	do isrc = 1, nsrc
		m_s_in%srcs(1,isrc)=m_s_in%srcs(1,isrc)+m_s_in%nsponge_x
		m_s_in%srcs(2,isrc)=m_s_in%srcs(2,isrc)+m_s_in%nsponge_z
	end do

	!validate virtual source and receiver locations
	call check_src_rec (m_s_in)

	! weights for FD scheme
	ax1 = m_s_in%dt*m_s_in%dt/(m_s_in%dx*m_s_in%dx)
	az1 = m_s_in%dt*m_s_in%dt/(m_s_in%dz*m_s_in%dz)
	ax2 = m_s_in%dt*m_s_in%dt/(m_s_in%dx*m_s_in%dx)
	az2 = m_s_in%dt*m_s_in%dt/(m_s_in%dz*m_s_in%dz)
	axz = m_s_in%dt*m_s_in%dt/(4.*m_s_in%dx*m_s_in%dz)
	apx1 = 1. / (2. * m_s_in%dx)
	apz1 = 1. / (2. * m_s_in%dz)
	asx1 = 1. / (2. * m_s_in%dx)
	asz1 = 1. / (2. * m_s_in%dz)

	! decon filter construction
	call from_param ("ncoeffs", ncoeffs, 5)
	call from_param ("eps", eps, 1e-20)

	call init_decon_filter (m_s_in, ncoeffs, eps)

end subroutine

subroutine init_decon_filter (m_s_in, ncoeffs, eps)
	type (model_type)	:: m_s_in
	integer				:: ncoeffs
	real				:: eps

	integer				:: stat, tmpsize
	real				:: expnorm
	real				:: ax_0, ax_1, ax_2, az_0, az_1, az_2
	real, allocatable, dimension (:)	:: tmp1, tmp2, crc, fcoeff

	! 4th order FD coefficients
	ax_2 = -1. / (m_s_in%dx * m_s_in%dx * 12.)
	ax_1 = -16. * ax_2
	ax_0 = 30. * ax_2
	az_2 = -1. / (m_s_in%dz * m_s_in%dz * 12.)
	az_1 = -16. * az_2
	az_0 = 30. * az_2

	! make 4th order filter
	m_s_in%dfilt = lapfac4 (eps, ax_0 + az_0, ax_1, az_1, ax_2, az_2, m_s_in%nx, ncoeffs)

	! check that the zero-lag coefficient is correct
	allocate (fcoeff (size(m_s_in%dfilt%flt) + 1))
	allocate (crc    (size (fcoeff)))
	
	fcoeff (1) = 1.
	fcoeff (2 : size (m_s_in%dfilt%flt) + 1) = m_s_in%dfilt%flt
	
	call correlate (fcoeff, fcoeff, crc, size(fcoeff))
	expnorm = (az_0 + ax_0 + eps) / crc(1)
	crc = crc * expnorm
	!write (0,*), "filter autocorrelation after norm = "
	!write (0,*), crc
	deallocate (crc)
	deallocate (fcoeff)

	! the decon result needs to be normalized, since the zero-lag cross-correlation element is an implied 1 in the polydiv code.
	! The normalization changes it so that the it comes out the same as if the zero-lag cross-correlation element was the central finite difference weight.

	tmpsize = ncoeffs * 2 + 1 ! an odd number, so that we'll have a central array element.
	call opolydiv_init (tmpsize, m_s_in%dfilt)

	allocate (tmp1 (tmpsize))
	allocate (tmp2 (tmpsize))
	tmp1 = .0
	tmp2 = .0

	tmp1 (tmpsize / 2 + 1) = 1.

	! run the decon on a spike
	stat = opolydiv_lop (.true.,  .false., tmp2, tmp1)
	stat = opolydiv_lop (.false., .false., tmp2, tmp1)

	! The normalization factor is the ratio between the deconvolved spike and the central FD weight.
	m_s_in%decon_norm = tmp1 (tmpsize / 2 + 1)  / (az_0 + ax_0 + eps)

	deallocate (tmp1)
	deallocate (tmp2)

	call opolydiv_close ()

end subroutine

subroutine check_src_rec (m_s_in)
	! check that all virtual sources/receivers are in the computational domain
	type(model_type)			:: m_s_in
	integer						:: irec, isrc, nrec_found, nsrc_found

	nrec_found=0
	nsrc_found=0

	! see how many receivers are in the computational domain
	nrec_found=0
	do irec = 1, size (m_s_in%recs,2)

		if (m_s_in%recs (1,irec) >= m_s_in%fx .and. &
			m_s_in%recs (1,irec) <= m_s_in%fx + m_s_in%nx - 1 .and. &
			m_s_in%recs (2,irec) >= m_s_in%fz .and. &
			m_s_in%recs (2,irec) <= m_s_in%fz + m_s_in%nz - 1) then

			nrec_found = nrec_found+1
		end if
	end do

	! see how many sources are in the computational domain
	nsrc_found=0
	do isrc = 1, size (m_s_in%srcs,2)

		if (m_s_in%srcs (1,isrc) >= m_s_in%fx .and. &
			m_s_in%srcs (1,isrc) <= m_s_in%fx + m_s_in%nx - 1 .and. &
			m_s_in%srcs (2,isrc) >= m_s_in%fz .and. &
			m_s_in%srcs (2,isrc) <= m_s_in%fz + m_s_in%nz - 1) then

			nsrc_found = nsrc_found+1
		end if
	end do

	if (nrec_found /= size (m_s_in%recs, 2) .or. nsrc_found /= size (m_s_in%srcs, 2)) then
		call seperr("Not all receivers/sources in computational domain")
	end if
end subroutine

subroutine make_wavelet (m_s_in)
	type(model_type)	:: m_s_in
	integer				:: ntw, i
	real				:: a, b, t2

	ntw = 1. / m_s_in%fpeak / m_s_in%dt
	allocate (m_s_in%wavelet (ntw*2 + 1))
	b = (3.1415926 * m_s_in%fpeak)**2
	a = 2. * b
	m_s_in%wavelet = 0
	do  i = 1, ntw+1
		t2 = ((i-1) * m_s_in%dt)**2
		m_s_in%wavelet (ntw+i)   = (1. - a * t2) * exp(-b * t2)
		m_s_in%wavelet (ntw+2-i) = m_s_in%wavelet (ntw+i)
	end do

end subroutine

subroutine global_init (m_s_in, d_s_in, io_s_in)
	type (model_type)	:: m_s_in
	type (distrib_type)	:: d_s_in
	type (io_type)		:: io_s_in

	m_s = m_s_in
	d_s = d_s_in
	io_s = io_s_in

end subroutine

! forward and adjoint operator
function L_oper (adj, add, m, d)  result(stat)
	integer					:: stat 
	logical, intent (in)	:: adj, add 
	real, dimension (:)		:: m, d 

	call stop_timer_num(tm_cgstep)

	call adjnull (adj, add, m, d)
	call prop_ts (adj, add, m, d)
	stat=0

	call start_timer_num(tm_cgstep)
end function

! data weighting
function W_oper (adj, add, m, d)  result(stat)
	integer					:: stat 
	logical, intent (in)	:: adj, add 
	real, dimension (:)		:: m, d 
	call adjnull (adj, add, m, d)
	call weight_geo_hydro (adj, add, m, d)
	stat=0
end function


subroutine prop_ts (adj, add, model, data)
	logical				:: adj, add
	real, dimension (:), target	:: model
	real, dimension (:), target	:: data

	integer				:: it, dt_print, i_print

	d_s%recdata => data
	d_s%srcmodel => model

	! zero all wavefields
	d_s%all = 0.

	if (adj) then
		write (0,*), "running adjoint operator"
	else
		write (0,*), "running forward operator"
	endif

	write (0,*), ""
	! for progress bar printing
	dt_print = m_s%nt / 10
	i_print = 0
	open (unit=0, carriagecontrol='fortran')
	call progressb (0)

	do it = 1, m_s%nt
		!============= ADJOINT OPERATOR (data ---> model) =====================
		! read from recorded data and generate source functions
		!======================================================================
		if (adj) then
			!read saved particle velocity fields and add to current particle velocity fields
			call start_timer_num(tm_read)
			call read_recs ("in_ux_data",.false., it, m_s, d_s, 1.,   1, iarray1=uxo_i, iarray_in=uxr_i)
			call read_recs ("in_uz_data",.false., it, m_s, d_s, 1.,   1, iarray1=uzo_i, iarray_in=uzr_i)
			call read_recs ("in_P_data", .false., it, m_s, d_s, apx1, 1, iarray1=uxo_i, iarray2=uzo_i, iarray_in=Pr_i)
			call stop_timer_num(tm_read)
	
			!advance velocities of velocity field
			call start_timer_num(tm_v4)
			call fdvsg3d_v2 (m_s%nx, m_s%nz, d_s%rho, d_s%mu, d_s%lambda, d_s%uxm, d_s%uxo, d_s%uxp, d_s%uzm, d_s%uzo, d_s%uzp, d_s%vxo, d_s%vzo, d_s%P, d_s%S, 1./m_s%dt, 1)
			call stop_timer_num(tm_v4)
	
			!do velocity boundary conditions on velocity field
			call start_timer_num(tm_sp)
			call fdvsg3d_sponge (d_s%uxo, m_s%sponge_x, m_s%sponge_z)
			call fdvsg3d_sponge (d_s%uxm, m_s%sponge_x, m_s%sponge_z)
			call fdvsg3d_sponge (d_s%uzo, m_s%sponge_x, m_s%sponge_z)
			call fdvsg3d_sponge (d_s%uzm, m_s%sponge_x, m_s%sponge_z)
			call stop_timer_num(tm_sp)
	
			call start_timer_num(tm_write)
			call write_srcs ("rec_uxb", .false., it, m_s, d_s, iarray=uxo_i, iarray_out=uxs_i)
			call write_srcs ("rec_uzb", .false., it, m_s, d_s, iarray=uzo_i, iarray_out=uzs_i)
			call stop_timer_num(tm_write)
			
		
		!============= FORWARD OPERATOR (model --> data) =====================
		! read from source functions and generate recorded data
		!=====================================================================
		else

			! forward modeling from point source
			if (m_s%proptype .eq. source .and. it .le. size (m_s%wavelet)) then
				if (m_s%isrc_type .eq. 0) then
					d_s%uxo (m_s%ixsrc+1, m_s%izsrc) = d_s%uxo (m_s%ixsrc+1, m_s%izsrc) +  apx1 * (d_s%lambda (m_s%ixsrc, m_s%izsrc) + d_s%mu (m_s%ixsrc, m_s%izsrc)) * m_s%wavelet (it)
					d_s%uxo (m_s%ixsrc-1, m_s%izsrc) = d_s%uxo (m_s%ixsrc-1, m_s%izsrc) -  apx1 * (d_s%lambda (m_s%ixsrc, m_s%izsrc) + d_s%mu (m_s%ixsrc, m_s%izsrc)) * m_s%wavelet (it)
					d_s%uzo (m_s%ixsrc, m_s%izsrc+1) = d_s%uzo (m_s%ixsrc, m_s%izsrc+1) +  apz1 * (d_s%lambda (m_s%ixsrc, m_s%izsrc) + d_s%mu (m_s%ixsrc, m_s%izsrc)) * m_s%wavelet (it)
					d_s%uzo (m_s%ixsrc, m_s%izsrc-1) = d_s%uzo (m_s%ixsrc, m_s%izsrc-1) -  apz1 * (d_s%lambda (m_s%ixsrc, m_s%izsrc) + d_s%mu (m_s%ixsrc, m_s%izsrc)) * m_s%wavelet (it)
	
				else if (m_s%isrc_type .eq. 1) then
					d_s%uxo (m_s%ixsrc, m_s%izsrc) = d_s%uxo (m_s%ixsrc, m_s%izsrc) +  d_s%rho (m_s%ixsrc, m_s%izsrc) * m_s%wavelet (it)
	
				else if (m_s%isrc_type .eq. 2) then
					d_s%uzo (m_s%ixsrc, m_s%izsrc) = d_s%uzo (m_s%ixsrc, m_s%izsrc) +  d_s%rho (m_s%ixsrc, m_s%izsrc) * m_s%wavelet (it)
	
				endif
	
			! forward modeling from virtual source functions
			else
				!add geophone trace data. The trace data must be interpolated to be in modeling dt intervals, otherwise we'll reach EOF.
				call start_timer_num(tm_read)
				call read_srcs ("in_ux_src", .false., it, m_s, d_s, 1., 1, iarray1=uxo_i, iarray_in=uxs_i)
				call read_srcs ("in_uz_src", .false., it, m_s, d_s, 1., 1, iarray1=uzo_i, iarray_in=uzs_i)
				call stop_timer_num(tm_read)
	
			endif
	
			!advance velocities of velocity field
			call start_timer_num(tm_v4)
			call fdvsg3d_v2 (m_s%nx, m_s%nz, d_s%rho, d_s%mu, d_s%lambda, d_s%uxm, d_s%uxo, d_s%uxp, d_s%uzm, d_s%uzo, d_s%uzp, d_s%vxo, d_s%vzo, d_s%P, d_s%S, 1./m_s%dt, 1)
			call stop_timer_num(tm_v4)

			!do velocity boundary conditions on velocity field
			call start_timer_num(tm_sp)
			call fdvsg3d_sponge (d_s%uxo, m_s%sponge_x, m_s%sponge_z)
			call fdvsg3d_sponge (d_s%uxm, m_s%sponge_x, m_s%sponge_z)
			call fdvsg3d_sponge (d_s%uzo, m_s%sponge_x, m_s%sponge_z)
			call fdvsg3d_sponge (d_s%uzm, m_s%sponge_x, m_s%sponge_z)
			call stop_timer_num(tm_sp)

			if (m_s%splitdisp .eq. 1 .and. (io_s%jrec == 1 .or. mod (it, io_s%jrec) == 1)) then
				call fdvsg3d_disp_space (m_s%nx, m_s%nz, d_s%uxo, d_s%uzo, d_s%Upx, d_s%Upz, d_s%Usx, d_s%Usz, m_s%dfilt, m_s%decon_norm)
			endif

			call start_timer_num(tm_write)
			call write_recs ("rec_ux", .false., it, m_s, d_s, iarray=uxo_i, iarray_out=uxr_i)
			call write_recs ("rec_uz", .false., it, m_s, d_s, iarray=uzo_i, iarray_out=uzr_i)
			call write_recs ("rec_P",  .false., it, m_s, d_s, iarray=P_i,   iarray_out=Pr_i)
			call stop_timer_num(tm_write)
	
		endif
		!write out wavefields to files
		call start_timer_num (tm_write)
		call write_out_files (it, m_s, d_s, io_s, adj)
		call stop_timer_num  (tm_write)

		! progress
		if ( mod (it, dt_print) == 0) then
			i_print = i_print + 1
			call progressb (i_print)
		endif

	end do
	write (0,*), ""

	! reverse time axis of virtual sources or receivers
	if (adj) then
		call reverse_srcs (m_s, d_s, uxs_i)
		call reverse_srcs (m_s, d_s, uzs_i)
	else
		call reverse_recs (m_s, d_s, uxr_i)
		call reverse_recs (m_s, d_s, uzr_i)
		call reverse_recs (m_s, d_s, Pr_i)
	endif

	return
end subroutine

! =============================================================
!     v2
!     update ux, uz, vy using 2nd order
! =============================================================

subroutine fdvsg3d_v2 (nx, nz, rho, mu, lambda, uxm, uxo, uxp, uzm, uzo, uzp, vxo, vzo, P, S, idt, timesign)
	integer					:: ix, iz, nx, nz
	real					:: dx, dz, idt
	real, dimension (:,:)	:: rho, mu, lambda
	real, dimension (:,:)	:: uxm, uxo, uxp, uzm, uzo, uzp, vxo, vzo, P, S
	integer					:: timesign
	real					:: cx1, cx2, cz1, cz2, cxz, cpx1, cpz1, csx1, csz1, murho, lam2mu, lammu

	murho  = mu (1, 1) * rho (1, 1)
	lam2mu = (lambda (1, 1) + 2.*mu (1, 1)) * rho (1, 1)
	lammu  = (lambda (1, 1) + mu (1, 1)) * rho (1, 1)
	cx1 = ax1 * lam2mu
	cz1 = az1 * lam2mu
	cx2 = ax2 * murho
	cz2 = az2 * murho
	cxz = axz * lammu

!$OMP PARALLEL private (iz, ix, murho, lam2mu, lammu)
!$OMP DO
	do iz = 2, nz-1
		do ix = 2, nx-1
			!murho  = mu (ix, iz) * rho (ix, iz)
			!lam2mu = (lambda (ix, iz) + 2.*mu (ix, iz)) * rho (ix, iz)
			!lammu  = (lambda (ix, iz) + mu (ix, iz)) * rho (ix, iz)
			!cx1 = ax1 * lam2mu
			!cz1 = az1 * lam2mu
			!cx2 = ax2 * murho
			!cz2 = az2 * murho
			!cxz = axz * lammu

			! spatial derivative in x
			uxp (ix, iz) =  cx1 * (uxo (ix-1, iz) - 2.*uxo (ix, iz) + uxo (ix+1, iz)) + &
							cxz * (uzo (ix-1, iz-1) - uzo (ix+1, iz-1) - uzo (ix-1, iz+1) + uzo (ix+1, iz+1)) + &
							cz2 * (uxo (ix, iz-1) -2.*uxo (ix, iz) + uxo (ix, iz+1)) + &
							2.*uxo (ix, iz) - uxm (ix, iz)

			! spatial derivative in z
			uzp (ix, iz) =	cz1 * (uzo (ix, iz-1) -2.*uzo (ix, iz) + uzo (ix, iz+1)) + &
							cxz * (uxo (ix-1, iz-1) - uxo (ix+1, iz-1) - uxo (ix-1, iz+1) + uxo (ix+1, iz+1)) + &
							cx2 * (uzo (ix-1, iz) -2.*uzo (ix, iz) + uzo (ix+1, iz)) + &
							2.*uzo (ix, iz) - uzm (ix, iz)

			vxo (ix, iz) = idt * (uxp (ix, iz) - uxo (ix, iz))
			vzo (ix, iz) = idt * (uzp (ix, iz) - uzo (ix, iz))

		end do
	end do
!$OMP END DO
!$OMP END PARALLEL


!$OMP PARALLEL private (iz, ix)
!$OMP DO
	do iz = 1, nz
		do ix = 1, nx
			uzm (ix, iz) = uzo (ix, iz)
			uzo (ix, iz) = uzp (ix, iz)
			uxm (ix, iz) = uxo (ix, iz)
			uxo (ix, iz) = uxp (ix, iz)
		end do
	end do
!$OMP END DO
!$OMP END PARALLEL

	!cpx1 = (lambda (1, 1) + mu (1, 1)) * apx1
 	!cpz1 = (lambda (1, 1) + mu (1, 1)) * apz1
	!csx1 = mu (1 ,1) * asx1
	!csz1 = mu (1 ,1) * asz1

!$OMP PARALLEL private (iz, ix)
!$OMP DO
  	do iz = 2, nz-1
   		do ix = 2, nx-1
  			P (ix, iz) = apx1 * (uxo (ix+1, iz) - uxo (ix-1, iz)) + apz1 * (uzo (ix, iz+1) - uzo (ix, iz-1))
  			S (ix, iz) = asx1 * (uxo (ix, iz+1) - uxo (ix, iz-1)) - asz1 * (uzo (ix+1, iz) - uzo (ix-1, iz))
  			!P (ix, iz) = (apx1 * (uxo (ix+1, iz) - uxo (ix-1, iz)) + apz1 * (uzo (ix, iz+1) - uzo (ix, iz-1))) * (lambda (ix, iz) + mu (ix, iz))
  			!S (ix, iz) = (asx1 * (uxo (ix, iz+1) - uxo (ix, iz-1)) - asz1 * (uzo (ix+1, iz) - uzo (ix-1, iz))) * mu (ix, iz)
   		end do
   	end do
!$OMP END DO
!$OMP END PARALLEL

end subroutine

! =============================================================
!	disp_space
!	do displacement decomposition in space domain.
! =============================================================

subroutine fdvsg3d_disp_space (nx, nz, ux, uz, Upx, Upz, Usx, Usz, dfilt, decon_norm)
	integer					:: nz, nx
	real, dimension (:,:)	:: ux, uz, Upx, Upz, Usx, Usz
	real					:: decon_norm
	type (filter)			:: dfilt

	real, save, allocatable, dimension (:)	:: ux_1d, uz_1d, tmp
	integer		:: i,k
	integer		:: stat
	real		:: Ux_d2x, Uz_d2z, Ux_d2z, Uz_d2x, Uz_dxz, Ux_dxz

	if (.not. allocated (ux_1d)) then
		allocate (ux_1d (nx * nz))
		allocate (uz_1d (nx * nz))
		allocate (tmp   (nx * nz))
		call opolydiv_init (nx * nz, dfilt)
		!write (0,*), "filter%flt =", dfilt%flt
	endif

	! integrate ux and uz to get ux and uz
!$OMP PARALLEL private (k, i)
!$OMP DO
	do k=1,nz
		do i=1,nx
			ux_1d ((k-1)*nx + i) = ux (i, k) ! put into 1d array for helical decon
			uz_1d ((k-1)*nx + i) = uz (i, k)
		end do
	end do
!$OMP END DO
!$OMP END PARALLEL

	! apply helical deconvolution of 4th order spatial derivative operator to ux and uz - this is equivalent to the normalization by |K| in wavenumber domain
	! For now, this is done only to one Y-plane of the 3D model ( j = iysrc ).  (wax-off, wax-on)
 	tmp = 0.
 	stat = opolydiv_lop (.true., .false.,  tmp, ux_1d) ! yes, we can certainly save here by reusing uz_1d. I'll do it later when I get the hang of things (promise)
 	stat = opolydiv_lop (.false., .false., tmp, ux_1d)
 	tmp = 0.
 	stat = opolydiv_lop (.true., .false.,  tmp, uz_1d)
 	stat = opolydiv_lop (.false., .false., tmp, uz_1d)

	uz_1d = uz_1d * decon_norm
	ux_1d = ux_1d * decon_norm

!$OMP PARALLEL private (k, i)
!$OMP DO
	do k = 3, nz-1
		do i = 2, nx-1

			Ux_d2x = ax1 * (ux_1d ((k-1)*nx + i-1) - 2.* ux_1d ((k-1)*nx + i) + ux_1d ((k-1)*nx + i+1))
			Uz_d2z = az1 * (uz_1d ((k-2)*nx + i  ) - 2.* uz_1d ((k-1)*nx + i) + uz_1d ((k  )*nx + i  ))
			
			Uz_d2x = ax2 * (uz_1d ((k-1)*nx + i-1) - 2.* uz_1d ((k-1)*nx + i) + uz_1d ((k-1)*nx + i+1))
			Ux_d2z = az2 * (ux_1d ((k-2)*nx + i  ) - 2.* ux_1d ((k-1)*nx + i) + ux_1d ((k  )*nx + i  ))

			Ux_dxz = axz * (ux_1d ((k-2)*nx + i-1) - ux_1d ((k-2)*nx + i+1) - ux_1d ((k  )*nx + i-1) + ux_1d ((k  )*nx + i+1))
			Uz_dxz = axz * (uz_1d ((k-2)*nx + i-1) - uz_1d ((k-2)*nx + i+1) - uz_1d ((k  )*nx + i-1) + uz_1d ((k  )*nx + i+1))

			Upx (i, k) = (Ux_d2x + Uz_dxz)
			Upz (i, k) = (Uz_d2z + Ux_dxz)
			Usx (i, k) = (Ux_d2z - Uz_dxz)
			Usz (i, k) = (Uz_d2x - Ux_dxz)
		end do
	end do
!$OMP END DO
!$OMP END PARALLEL

end subroutine

! Weight the data residual, so as not to include the hydrophone data in the inversion (for the moment)
subroutine weight_geo_hydro (adj, add, model, data)
	logical				:: adj, add
	real, dimension (:), target	:: model
	real, dimension (:), target	:: data
	integer						:: nrecs, nsrcs

	nrecs = size (m_s%recs,2)
	nsrcs = size (m_s%srcs,2)

	! make hydrophone data = 0, so that it doesn't get into the model gradient calculation
	if (adj) then
		model (1 + nrecs * m_s%nt * uxr_i : nrecs * m_s%nt * (uxr_i + 1)) = 1. *  data (1 + nrecs * m_s%nt * uxr_i : nrecs * m_s%nt * (uxr_i + 1))
		model (1 + nrecs * m_s%nt * uzr_i : nrecs * m_s%nt * (uzr_i + 1)) = 1. *  data (1 + nrecs * m_s%nt * uzr_i : nrecs * m_s%nt * (uzr_i + 1))
		model (1 + nrecs * m_s%nt * Pr_i  : nrecs * m_s%nt * (Pr_i  + 1)) = 0. *  data (1 + nrecs * m_s%nt * Pr_i  : nrecs * m_s%nt * (Pr_i  + 1))
	else
	! forward: both model and data are of data dimensions. I want to not use the hydrophone at the moment, so it is given a weight of 0.
		data (1 + nrecs * m_s%nt * uxr_i : nrecs * m_s%nt * (uxr_i + 1)) = 1. *  model (1 + nrecs * m_s%nt * uxr_i : nrecs * m_s%nt * (uxr_i + 1))
		data (1 + nrecs * m_s%nt * uzr_i : nrecs * m_s%nt * (uzr_i + 1)) = 1. *  model (1 + nrecs * m_s%nt * uzr_i : nrecs * m_s%nt * (uzr_i + 1))
		data (1 + nrecs * m_s%nt * Pr_i  : nrecs * m_s%nt * (Pr_i  + 1)) = 0. *  model (1 + nrecs * m_s%nt * Pr_i  : nrecs * m_s%nt * (Pr_i  + 1))

	endif

end subroutine
subroutine fdvsg3d_sponge(array, xsponge, zsponge)
	real :: array(:,:)
	real :: xsponge (:), zsponge (:)

	call fdvsg3d_vabs_z_l(array, zsponge)

	call fdvsg3d_vabs_z_u(array, zsponge)

	call fdvsg3d_vabs_x_l(array, xsponge)

	call fdvsg3d_vabs_x_u(array, xsponge)
end subroutine

! ======================================================================
!     vabs_x_l
!     dump velocity near top boundary
!     using Cerjan et al. sponge absorbing boundary conditions
!     (see also mksponge)
! ======================================================================

subroutine fdvsg3d_vabs_x_l(array, xsponge)
	real :: array(:,:)
	real :: xsponge (:)
	integer :: ix,iz

!$OMP PARALLEL private (iz, ix)
!$OMP DO
	do iz=1,m_s%nz
		do ix=1,m_s%nsponge_x
			array (ix,iz)=array(ix,iz)*xsponge(ix)
		end do
	end do
!$OMP END DO
!$OMP END PARALLEL
end subroutine



! ======================================================================
!     vabs_bo
!     dump velocity near bottom boundary
!     using Cerjan et al. sponge absorbing boundary conditions
!     (see also mksponge)
! ======================================================================
subroutine fdvsg3d_vabs_x_u(array, xsponge)
	real :: array(:,:)
	real :: xsponge (:)
	integer :: iz,ix


!$OMP PARALLEL private (iz, ix)
!$OMP DO
	do iz=1,m_s%nz
		do ix=1,m_s%nsponge_x
			array(m_s%nx+1-ix,iz)=array(m_s%nx+1-ix,iz)*xsponge(ix)
		end do
	end do
!$OMP END DO
!$OMP END PARALLEL
end subroutine

! ======================================================================
!     vabs_z_l
!     dump velocity near top boundary
!     using Cerjan et al. sponge absorbing boundary conditions
!     (see also mksponge)
! ======================================================================

subroutine fdvsg3d_vabs_z_l(array, zsponge)
	integer :: iz,ix
	real :: array(:,:)
	real :: zsponge (:)
	integer, external	:: omp_get_thread_num


!$OMP PARALLEL private (iz, ix)
!$OMP DO
	do iz=1,m_s%nsponge_z
		do ix=1+m_s%nsponge_x,m_s%nx-m_s%nsponge_x
			array(ix,iz)=array(ix,iz)*zsponge(iz)
		end do
	end do
!$OMP END DO
!$OMP END PARALLEL




end subroutine



! ======================================================================
!     vabs_bo
!     dump velocity near bottom boundary
!     using Cerjan et al. sponge absorbing boundary conditions
!     (see also mksponge)
! ======================================================================
subroutine fdvsg3d_vabs_z_u(array, zsponge)
	real :: array(:,:)
	real :: zsponge (:)
	integer :: iz,ix


!$OMP PARALLEL private (iz, ix)
!$OMP DO
	do iz=1,m_s%nsponge_z
		do ix=1+m_s%nsponge_x,m_s%nx-m_s%nsponge_x
			array(ix,m_s%nz+1-iz)=array(ix,m_s%nz+1-iz)*zsponge(iz)
		end do
	end do
!$OMP END DO
!$OMP END PARALLEL
end subroutine

subroutine mk_sponge (m_s_in)
	type (model_type)	:: m_s_in
	real				:: a0, damper
	integer				:: i, imax_x, imax_z, xsponge, zsponge
	real				:: pi = 4. * atan(1.)

	call from_param ("zsponge", zsponge, 10)
	call from_param ("xsponge", xsponge, 10)
	call from_param ("damper",  damper, 0.6)

	imax_z = zsponge
	a0 = damper / real (imax_z)

	allocate (m_s_in%sponge_z (imax_z))
	m_s_in%nsponge_z = imax_z
	do i = 1, imax_z
		m_s_in%sponge_z (i) = exp (- (a0 * real (imax_z + 1 - i))**2)
		!m_s_in%sponge_z (i) = 1. - damper * (1. + cos (pi * i / (imax_z + 1)))
	end do

	imax_x = xsponge
	a0 = damper / real (imax_x)

	allocate (m_s_in%sponge_x (imax_x))
	m_s_in%nsponge_x = imax_x
	do i = 1, imax_x
		m_s_in%sponge_x (i) = exp(- (a0 * real (imax_x + 1 - i))**2)
		!sponge_x (i) = 1. - damper * (1. + cos (pi * i / (imax_x + 1)))
	end do

	!write (0,*), "sponge_x:", m_s_in%sponge_x
	!write (0,*), "sponge_z:", m_s_in%sponge_z
end subroutine

subroutine progressb (j)
	integer	:: j, k
	character (len=12) :: bar="|          |"

	if (j <= 10) then
		do k = 1, j
			bar (1+k:1+k) = "*"
		enddo
		do k = j+1, 10
			bar (1+k:1+k) = " "
		enddo
		write (unit=0, fmt="(a1,a1,a12)") '+', char(13), bar
	endif

	return
end subroutine

integer function pad2 (n)
	integer	:: n
	pad2 = 1

	do while (pad2 .lt. n)
		pad2 = pad2*2
	end do

	return
end function pad2


end module
